# -*- coding: utf-8 -*-
# 'future imports'
from __future__ import unicode_literals

# 'Standard Library'
import json

from calendar import timegm

try:
    from rfc822 import parsedate
except ImportError:
    from email.utils import parsedate


class TwitterModel(object):

    """ Base class from which all twitter models will inherit. """

    def __init__(self, **kwargs):
        self.param_defaults = {}

    def __str__(self):
        """ Returns a string representation of TwitterModel. By default
        this is the same as AsJsonString(). """
        return self.AsJsonString()

    def __eq__(self, other):
        return other and self.AsDict() == other.AsDict()

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        if hasattr(self, "id"):
            return hash(self.id)
        else:
            raise TypeError(
                "unhashable type: {} (no id attribute)".format(type(self))
            )

    def AsJsonString(self, ensure_ascii=True):
        """ Returns the TwitterModel as a JSON string based on key/value
        pairs returned from the AsDict() method. """
        return json.dumps(
            self.AsDict(), ensure_ascii=ensure_ascii, sort_keys=True
        )

    def AsDict(self):
        """ Create a dictionary representation of the object. Please see inline
        comments on construction when dictionaries contain TwitterModels. """
        data = {}

        for (key, value) in self.param_defaults.items():

            # If the value is a list, we need to create a list to hold the
            # dicts created by an object supporting the AsDict() method,
            # i.e., if it inherits from TwitterModel. If the item in the list
            # doesn't support the AsDict() method, then we assign the value
            # directly. An example being a list of Media objects contained
            # within a Status object.
            if isinstance(getattr(self, key, None), (list, tuple, set)):
                data[key] = list()
                for subobj in getattr(self, key, None):
                    if getattr(subobj, "AsDict", None):
                        data[key].append(subobj.AsDict())
                    else:
                        data[key].append(subobj)

            # Not a list, *but still a subclass of TwitterModel* and
            # and we can assign the data[key] directly with the AsDict()
            # method of the object. An example being a Status object contained
            # within a User object.
            elif getattr(getattr(self, key, None), "AsDict", None):
                data[key] = getattr(self, key).AsDict()

            # If the value doesn't have an AsDict() method, i.e., it's not
            # something that subclasses TwitterModel, then we can use direct
            # assigment.
            elif getattr(self, key, None):
                data[key] = getattr(self, key, None)
        return data

    @classmethod
    def NewFromJsonDict(cls, data, **kwargs):
        """ Create a new instance based on a JSON dict. Any kwargs should be
        supplied by the inherited, calling class.

        Args:
            data: A JSON dict, as converted from the JSON in the twitter API.

        """

        json_data = data.copy()
        if kwargs:
            for key, val in kwargs.items():
                json_data[key] = val

        c = cls(**json_data)
        c._json = data
        return c


class Media(TwitterModel):

    """A class representing the Media component of a tweet. """

    def __init__(self, **kwargs):
        self.param_defaults = {
            "display_url": None,
            "expanded_url": None,
            "ext_alt_text": None,
            "id": None,
            "media_url": None,
            "media_url_https": None,
            "sizes": None,
            "type": None,
            "url": None,
            "video_info": None,
        }

        for (param, default) in self.param_defaults.items():
            setattr(self, param, kwargs.get(param, default))

    def __repr__(self):
        return "Media(ID={media_id}, Type={media_type}, DisplayURL='{url}')".format(
            media_id=self.id, media_type=self.type, url=self.display_url
        )


class List(TwitterModel):

    """A class representing the List structure used by the twitter API. """

    def __init__(self, **kwargs):
        self.param_defaults = {
            "description": None,
            "following": None,
            "full_name": None,
            "id": None,
            "member_count": None,
            "mode": None,
            "name": None,
            "slug": None,
            "subscriber_count": None,
            "uri": None,
            "user": None,
        }

        for (param, default) in self.param_defaults.items():
            setattr(self, param, kwargs.get(param, default))

        if "user" in kwargs:
            self.user = User.NewFromJsonDict(kwargs.get("user"))

    def __repr__(self):
        return "List(ID={list_id}, FullName={full_name!r}, Slug={slug}, User={user})".format(
            list_id=self.id,
            full_name=self.full_name,
            slug=self.slug,
            user=self.user.screen_name,
        )


class Category(TwitterModel):

    """A class representing the suggested user category structure. """

    def __init__(self, **kwargs):
        self.param_defaults = {
            "name": None,
            "size": None,
            "slug": None,
        }

        for (param, default) in self.param_defaults.items():
            setattr(self, param, kwargs.get(param, default))

    def __repr__(self):
        return "Category(Name={name!r}, Slug={slug}, Size={size})".format(
            name=self.name, slug=self.slug, size=self.size
        )


class DirectMessage(TwitterModel):

    """A class representing a Direct Message. """

    def __init__(self, **kwargs):
        self.param_defaults = {
            "created_at": None,
            "id": None,
            "recipient_id": None,
            "sender_id": None,
            "text": None,
        }

        for (param, default) in self.param_defaults.items():
            setattr(self, param, kwargs.get(param, default))

    def __repr__(self):
        if self.text and len(self.text) > 140:
            text = "{text}[...]".format(text=self.text[:140])
        else:
            text = self.text
        return "DirectMessage(ID={dm_id}, Sender={sender}, Created={time}, Text='{text!r}')".format(
            dm_id=self.id,
            sender=self.sender_id,
            time=self.created_at,
            text=text,
        )


class Trend(TwitterModel):

    """ A class representing a trending topic. """

    def __init__(self, **kwargs):
        self.param_defaults = {
            "events": None,
            "name": None,
            "promoted_content": None,
            "query": None,
            "timestamp": None,
            "url": None,
            "tweet_volume": None,
        }

        for (param, default) in self.param_defaults.items():
            setattr(self, param, kwargs.get(param, default))

    def __repr__(self):
        return "Trend(Name={0!r}, Time={1}, URL={2})".format(
            self.name, self.timestamp, self.url
        )

    @property
    def volume(self):
        return self.tweet_volume


class Hashtag(TwitterModel):

    """ A class representing a twitter hashtag. """

    def __init__(self, **kwargs):
        self.param_defaults = {"text": None}

        for (param, default) in self.param_defaults.items():
            setattr(self, param, kwargs.get(param, default))

    def __repr__(self):
        return "Hashtag(Text={text!r})".format(text=self.text)


class Url(TwitterModel):

    """ A class representing an URL contained in a tweet. """

    def __init__(self, **kwargs):
        self.param_defaults = {"expanded_url": None, "url": None}

        for (param, default) in self.param_defaults.items():
            setattr(self, param, kwargs.get(param, default))

    def __repr__(self):
        return "URL(URL={url}, ExpandedURL={eurl})".format(
            url=self.url, eurl=self.expanded_url
        )


class UserStatus(TwitterModel):

    """ A class representing the UserStatus structure. This is an abbreviated
    form of the twitter.User object. """

    _connections = {
        "following": False,
        "followed_by": False,
        "following_received": False,
        "following_requested": False,
        "blocking": False,
        "muting": False,
    }

    def __init__(self, **kwargs):
        self.param_defaults = {
            "blocking": False,
            "followed_by": False,
            "following": False,
            "following_received": False,
            "following_requested": False,
            "id": None,
            "id_str": None,
            "muting": False,
            "name": None,
            "screen_name": None,
        }

        for (param, default) in self.param_defaults.items():
            setattr(self, param, kwargs.get(param, default))

        if "connections" in kwargs:
            for param in self._connections:
                if param in kwargs["connections"]:
                    setattr(self, param, True)

    @property
    def connections(self):
        return {
            "following": self.following,
            "followed_by": self.followed_by,
            "following_received": self.following_received,
            "following_requested": self.following_requested,
            "blocking": self.blocking,
            "muting": self.muting,
        }

    def __repr__(self):
        connections = [
            param for param in self.connections if getattr(self, param)
        ]
        return "UserStatus(ID={uid}, ScreenName={sn}, Connections=[{conn}])".format(
            uid=self.id, sn=self.screen_name, conn=", ".join(connections)
        )


class User(TwitterModel):

    """A class representing the User structure. """

    def __init__(self, **kwargs):
        self.param_defaults = {
            "contributors_enabled": None,
            "created_at": None,
            "default_profile": None,
            "default_profile_image": None,
            "description": None,
            "email": None,
            "favourites_count": None,
            "followers_count": None,
            "following": None,
            "friends_count": None,
            "geo_enabled": None,
            "id": None,
            "id_str": None,
            "lang": None,
            "listed_count": None,
            "location": None,
            "name": None,
            "notifications": None,
            "profile_background_color": None,
            "profile_background_image_url": None,
            "profile_background_image_url_https": None,
            "profile_background_tile": None,
            "profile_banner_url": None,
            "profile_image_url": None,
            "profile_image_url_https": None,
            "profile_link_color": None,
            "profile_sidebar_border_color": None,
            "profile_sidebar_fill_color": None,
            "profile_text_color": None,
            "profile_use_background_image": None,
            "protected": None,
            "screen_name": None,
            "status": None,
            "statuses_count": None,
            "time_zone": None,
            "url": None,
            "utc_offset": None,
            "verified": None,
            "withheld_in_countries": None,
            "withheld_scope": None,
        }

        for (param, default) in self.param_defaults.items():
            setattr(self, param, kwargs.get(param, default))

    def __repr__(self):
        return "User(ID={uid}, ScreenName={sn})".format(
            uid=self.id, sn=self.screen_name
        )

    @classmethod
    def NewFromJsonDict(cls, data, **kwargs):
        from twitter import Status

        if data.get("status", None):
            status = Status.NewFromJsonDict(data.get("status"))
            return super(cls, cls).NewFromJsonDict(data=data, status=status)
        else:
            return super(cls, cls).NewFromJsonDict(data=data)


class Status(TwitterModel):
    """A class representing the Status structure used by the twitter API.
    """

    def __init__(self, **kwargs):
        self.param_defaults = {
            "contributors": None,
            "coordinates": None,
            "created_at": None,
            "current_user_retweet": None,
            "favorite_count": None,
            "favorited": None,
            "full_text": None,
            "geo": None,
            "hashtags": None,
            "id": None,
            "id_str": None,
            "in_reply_to_screen_name": None,
            "in_reply_to_status_id": None,
            "in_reply_to_user_id": None,
            "lang": None,
            "location": None,
            "media": None,
            "place": None,
            "possibly_sensitive": None,
            "quoted_status": None,
            "quoted_status_id": None,
            "quoted_status_id_str": None,
            "retweet_count": None,
            "retweeted": None,
            "retweeted_status": None,
            "scopes": None,
            "source": None,
            "text": None,
            "truncated": None,
            "urls": None,
            "user": None,
            "user_mentions": None,
            "withheld_copyright": None,
            "withheld_in_countries": None,
            "withheld_scope": None,
        }

        for (param, default) in self.param_defaults.items():
            setattr(self, param, kwargs.get(param, default))

        if kwargs.get("full_text", None):
            self.tweet_mode = "extended"
        else:
            self.tweet_mode = "compatibility"

    @property
    def created_at_in_seconds(self):
        """ Get the time this status message was posted, in seconds since
        the epoch (1 Jan 1970).

        Returns:
            int: The time this status message was posted, in seconds since
            the epoch.
        """
        return timegm(parsedate(self.created_at))

    def __repr__(self):
        """ A string representation of this twitter.Status instance.
        The return value is the ID of status, username and datetime.

        Returns:
            string: A string representation of this twitter.Status instance with
            the ID of status, username and datetime.
        """
        if self.tweet_mode == "extended":
            text = self.full_text
        else:
            text = self.text
        if self.user:
            return "Status(ID={0}, ScreenName={1}, Created={2}, Text={3!r})".format(
                self.id, self.user.screen_name, self.created_at, text
            )
        else:
            return "Status(ID={0}, Created={1}, Text={2!r})".format(
                self.id, self.created_at, text
            )

    @classmethod
    def NewFromJsonDict(cls, data, **kwargs):
        """ Create a new instance based on a JSON dict.

        Args:
            data: A JSON dict, as converted from the JSON in the twitter API

        Returns:
            A twitter.Status instance
        """
        current_user_retweet = None
        hashtags = None
        media = None
        quoted_status = None
        retweeted_status = None
        urls = None
        user = None
        user_mentions = None

        # for loading extended tweets from the streaming API.
        if "extended_tweet" in data:
            for k, v in data["extended_tweet"].items():
                data[k] = v

        if "user" in data:
            user = User.NewFromJsonDict(data["user"])
        if "retweeted_status" in data:
            retweeted_status = Status.NewFromJsonDict(data["retweeted_status"])
        if "current_user_retweet" in data:
            current_user_retweet = data["current_user_retweet"]["id"]
        if "quoted_status" in data:
            quoted_status = Status.NewFromJsonDict(data.get("quoted_status"))

        if "entities" in data:
            if "urls" in data["entities"]:
                urls = [
                    Url.NewFromJsonDict(u) for u in data["entities"]["urls"]
                ]
            if "user_mentions" in data["entities"]:
                user_mentions = [
                    User.NewFromJsonDict(u)
                    for u in data["entities"]["user_mentions"]
                ]
            if "hashtags" in data["entities"]:
                hashtags = [
                    Hashtag.NewFromJsonDict(h)
                    for h in data["entities"]["hashtags"]
                ]
            if "media" in data["entities"]:
                media = [
                    Media.NewFromJsonDict(m) for m in data["entities"]["media"]
                ]

        # the new extended entities
        if "extended_entities" in data:
            if "media" in data["extended_entities"]:
                media = [
                    Media.NewFromJsonDict(m)
                    for m in data["extended_entities"]["media"]
                ]

        return super(cls, cls).NewFromJsonDict(
            data=data,
            current_user_retweet=current_user_retweet,
            hashtags=hashtags,
            media=media,
            quoted_status=quoted_status,
            retweeted_status=retweeted_status,
            urls=urls,
            user=user,
            user_mentions=user_mentions,
        )



# -*- coding: utf-8 -*-
""" Part of the `AutoSys` package
        copyright (c) 2019 Michael Treanor
        https://www.github.com/skeptycal/autosys
        https://www.twitter.com/skeptycal

    `AutoSys` is licensed under the `MIT License
        `<https://opensource.org/licenses/MIT>`
    """
__license__ = "MIT"

# 'Standard Library'
import os
import re
import sys

from os import linesep as NL

# !---------------------------------------------- String Utilities


def split_it(s: str, delimiter: str = "\s"):
    return re.split(pattern=s, string=delimiter)


def arg_str(*args, sep=" ") -> (str):
    """ Return one string created from *args sequence.

        sep - string used in between joined elements.

        each element is converted to a string where possible, otherwise
        it is ignored and silently logged as an error
        """
    from typing import Any, List

    _: List[Any] = []
    for arg in args:
        try:  # convert args to strings
            _.append(str(arg))
        except:
            try:
                log.error(
                    f"An argument could not be converted to a string for display - {arg}"
                )
            except:
                pass  # skip debug output if not available
    return sep.join(_)


def brerr(n: int = 1) -> (int):
    """ Yes, this is a CLI version of a <br /> tag ... sent to STDERR.

        n : int - number of line breaks to print

        return int 0 for success else 1
        """
    try:
        print("\n" * n, end="", sep="", file=sys.stderr)
        return 0
    except:
        return 1



# 'Standard Library'
from pathlib import Path
from sys import argv

from typing import Any, Iterable, List


def LOG_PATH(f: str = __name__) -> Path:
    return Path().cwd().resolve() / f


def export(fn: Any) -> Any:
    """ Decorator to export functions and classes. """
    from sys import modules

    mod: Any = modules[fn.__module__]
    if hasattr(mod, "__all__"):
        mod.__all__.append(fn.__name__)
    else:
        mod.__all__ = [fn.__name__]
    return fn


def filter_list(
    d: Iterable = dir(),
    prefix: str = "_",
    whitelist: Iterable = [],
    blacklist: Iterable = [],
) -> List:
    """ Filter a list with prefix to exclude, whitelist, and blacklist. """
    return [
        x
        for x in sorted(d)
        if x in whitelist
        or ((not x.startswith(prefix) and x not in blacklist))
    ]


_EXPORT_BLACKLIST: str = ["arepl_store", "howdoi", "help"]
_EXPORT_WHITELIST: str = ["_debug_", "_log_flag_", "_verbose_"]


def all_export(
    d: Iterable = dir(),
    prefix: str = "_",
    whitelist: Iterable = _EXPORT_WHITELIST,
    blacklist: Iterable = _EXPORT_BLACKLIST,
) -> List:
    """ Return a list of all exports not starting with `_` """
    return filter_list(
        d=d, prefix=prefix, whitelist=whitelist, blacklist=blacklist
    )


__all__ = all_export()

if __name__ == "__main__":
    print(__all__)



# 'Standard Library'
import os
import tempfile

# 'package imports'
import contextlib2 as contextlib
import pathlib2 as pathlib
import temporary.util


@contextlib.contextmanager
def temp_file(content=None, suffix="", prefix="tmp", parent_dir=None):
    """
    Create a temporary file and optionally populate it with content. The file
    is deleted when the context exits.

    The temporary file is created when entering the context manager and
    deleted when exiting it.
    >>> import temporary
    >>> with temporary.temp_file() as temp_file:
    ...     assert temp_file.exists()
    >>> assert not temp_file.exists()

    The user may also supply some content for the file to be populated with:
    >>> with temporary.temp_file('hello!') as temp_file:
    ...     with temp_file.open() as f:
    ...         assert f.read() == 'hello!'

    The temporary file can be placed in a custom directory:
    >>> with temporary.temp_dir() as temp_dir:
    ...     with temporary.temp_file(parent_dir=temp_dir) as temp_file:
    ...         assert temp_file.parent == temp_dir

    If, for some reason, the user wants to delete the temporary file before
    exiting the context, that's okay too:
    >>> with temporary.temp_file() as temp_file:
    ...     temp_file.unlink()
    """
    binary = isinstance(content, (bytes, bytearray))
    parent_dir = parent_dir if parent_dir is None else str(parent_dir)
    fd, abs_path = tempfile.mkstemp(suffix, prefix, parent_dir, text=False)
    path = pathlib.Path(abs_path)
    try:
        try:
            if content:
                os.write(fd, content if binary else content.encode())
        finally:
            os.close(fd)
        yield path.resolve()
    finally:
        with temporary.util.allow_missing_file():
            path.unlink()



"""
ref: https://stackoverflow.com/a/56709915/9878098
I've found del to be useful for pseudo-manual memory management when handling large data with Numpy. For example:
"""

for image_name in large_image_set:
    large_image = io.imread(image_name)
    height, width, depth = large_image.shape

    large_mask = np.all(large_image == "<some_condition>")

    # Clear memory, make space
    del large_image
    gc.collect()

    large_processed_image = np.zeros((height, width, depth))
    large_processed_image[large_mask] = new_value
    io.imsave("processed_image.png", large_processed_image)

    # Clear memory, make space
    del large_mask, large_processed_image
    gc.collect()

"""
This can be the difference between bringing a script to a grinding halt as the system swaps like mad when the Python GC can't keep up, and it running perfectly smooth below a loose memory threshold that leaves plenty of headroom to use the machine to browse and code while it's working.
"""



#!/usr/bin/env python3
# -*- coding: UTF-8 -*-

# 'Standard Library'
import re

regex = r"\A[\w\s:\-<>\(\)\+\*]+\Z"
regex = r"\A[\w\s:\-<>\(\)\+\*\&\@]+\Z"
regex = r"[^\w\s\-\(\)\>\^\<\&\@\!\:\+\=\*]"

REGEX_SAFE_STRING = r"[^\w\s\-\(\)\>\^\<\&\@\!\:\+\=\*\'\"]"
RE_SAFE_STRING = re.compile(REGEX_SAFE_STRING, re.IGNORECASE | re.MULTILINE)
# https://regex101.com/r/9YvEas/1

RE_FLAGS_DEBUG = re.IGNORECASE | re.MULTILINE | re.VERBOSE | re.DEBUG

test_str = ("def fibonacci_2(n: int) -> int:\n"
            "    if n > 1:\n"
            "        return fibonacci(n - 1) + fibonacci(n - 2)\n"
            "    else:\n"
            "        return n=^&*(@@!)")

matches = re.search(regex, test_str,
                    re.IGNORECASE | re.MULTILINE | re.VERBOSE | re.DEBUG)


def is_safe(s) -> (bool):
    return re.search(RE_SAFE_STRING, s,
                     re.IGNORECASE | re.MULTILINE | re.VERBOSE | re.DEBUG)


print(matches)
# if matches:
#     print("Match was found at {start}-{end}: {match}".format(
#         start=matches.start(), end=matches.end(), match=matches.group()))

#     for groupNum in range(0, len(matches.groups())):
#         groupNum = groupNum + 1

#         print("Group {groupNum} found at {start}-{end}: {group}".format(groupNum=groupNum,
#                                                                         start=matches.start(groupNum), end=matches.end(groupNum), group=matches.group(groupNum)))

# Note: for Python 2.7 compatibility, use ur"" to prefix the regex and u"" to prefix the test string and substitution.



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

if True:  # !------------------------ Package Imports
    from enum import unique, Enum, IntEnum
    from typing import Dict, Iterable, Tuple
    from time import sleep
    from dataclasses import dataclass

    import numpy as np

    from autosys import *
    from autosys.debug import SUPPORTS_COLOR

if True:  # !------------------------ Constants
    @unique
    class NpAlign(Enum):
        left: str = "<"
        right: str = ">"
        center: str = "^"

if True:  # !------------------------ ANSI Colors
    @unique
    class FB(Enum):
        fg: int = 3
        bg: int = 4


@dataclass(frozen=True)
class Code8(Enum,str):
    black: str = "0"
    red: str = "1"
    green: str = "2"
    yellow: str = "3"
    blue: str = "4"
    magenta: str = "5"
    cyan: str = "6"
    white: str = "7"


@dataclass(frozen=True)
class Code16:
    black: str = "0"
    red: str = "1"
    green: str = "2"
    yellow: str = "3"
    blue: str = "4"
    magenta: str = "5"
    cyan: str = "6"
    white: str = "7"
    bright_black: str = "0;1"
    bright_red: str = "1;1"
    bright_green: str = "2;1"
    bright_yellow: str = "3;1"
    bright_blue: str = "4;1"
    bright_magenta: str = "5;1"
    bright_cyan: str = "6;1"
    bright_white: str = "7;1"


@dataclass(frozen=True)
class Effects:
    bold: str = "\u001b[1m"
    underline: str = "\u001b[4m"
    reverse: str = "\u001b[7m"
    reset: str = "\u001b[0m"


@dataclass()
class Ansi(str):
    reset: str = "\u001b[0m"

    def __init__(self, *args, **kwargs):
        super(*args, **kwargs)

    def _color8(self, c: Code8 = "1", fb: FB = FB.fg):
        """ get either fg or bg color (8) - default fg """
        return f"{self}\u001b[{fb}{c}m"

    def _color16(self, c: Code16 = Code16.blue, fb: FB = FB.fg):
        """ get either fg or bg color (16) - default fg """
        return f"{self}\u001b[{fb}{c}m"

    def _color256(self, c="229", fb: FB = FB.fg):
        """ get either fg or bg color (256) - default fg """
        return f"\u001b[{fb}8;5;{c}m"

    def fb(self, fg="0", bg="229"):
        """ <F>oreground / <B>ackground 256 color combo """
        return f"\u001b[38;5;{fg}m\u001b[48;5;{bg}m"

a = Ansi()
DEFAULT_NORMAL = a.fb(fg=2, bg=229)
DEFAULT_REVERSE = a.fb(fg=16, bg=229)

class ColorToggle:
    norm: str  = DEFAULT_NORMAL,
    rev: str = DEFAULT_REVERSE,
    color_rows: bool = True,
    color_cols: bool = True,
    do_checkerboard: bool = True
    on: bool = True
    _row_state: bool = True
    _col_state: bool = True


    """ use T/F array of any shape as a color pattern?

        e.g.
        pattern: nparray() = [
            [x o x],
            [o x o],
            [x o x]
        ]
        """

    reset: str = "\u001b[0m"

    # try dataclass init ...

    # def __init__(
        #     self,
        #     norm=DEFAULT_NORMAL,
        #     rev=DEFAULT_REVERSE,
        #     rows=True,
        #     cols=True,
        #     checkerboard=True,
        # ):
        # # set colors
        # self._normal = norm
        # self._reverse = rev

        # # set features
        # self._on: bool = True  # turn all colors on and off
        # self._rows: bool = rows  # color rows
        # self._cols: bool = cols  # color columns
        # self._checkerboard: bool = checkerboard  # do checkerboard?

        # # store initial state
        # self._row_state: bool = self._rows  # state of rows
        # self._col_state: bool = self._cols  # state of columns

    @property
    def row(self):
        self._row_state = not self._row_state
        # if self._checkerboard:
        #     return self.col
        return self

    @property
    def col(self):
        self._col_state = not self._col_state
        return self

    def __str__(self):
        return self._row_state and self._normal or self._reverse


class NpChart:
    a: Iterable
    size: int = 4
    padding: int = 1
    pad: str = " "
    align: NpAlign = NpAlign.left



def np_row(
    a: Iterable,
    size: int = 4,
    padding: int = 1,
    pad: str = " ",
    color_key: Tuple = (True,True, True, True),
    align: NpAlign = NpAlign.left,
):
    """ Format a row of data for cli output.

        param       description                             default
        ===========================================================
        a -         any iterable data set
        size -      space allowed for each data point       4
        padding -   space between each data point           1
        pad -       character used for padding              ' '
        alt_color - should it alternate colors?             False
        align -     alignment of data in cells              '<'

        color_tuple - options for coloring the chart
            color: bool = True,
            alt_rows: bool = True,
            alt_columns: bool = True,
            checkerboard: bool = True,
        """

    ct = ColorToggle() if SUPPORTS_COLOR else ""
    return f'{ct.row}{"".join([f"{ct.col}{x:{align}{size}}{pad*padding}" for x in a])}{ct.reset}'


a = np.arange(5)
arithmetic_lst = ["a", "a + 10", "a - 10", "a * 2", "a / 2", "a // 2"]
for arithmetic in arithmetic_lst:
    evl = eval(arithmetic)
    fmt = np_row(evl, 3, align=NpAlign.right)
    print(f"{arithmetic:>10}: {fmt}")
print()


# print methods of numpy.ndarray without dunders
# print([x for x in dir(evl) if not x.startswith("_")])



#!/usr/bin/env python3
# -*- coding: UTF-8 -*-
"""
    version.py - Version and Package Information

    Conforms to PEP 566 Metadata for Python Software Packages 2.1
    https://www.python.org/dev/peps/pep-0566/

    - Core metadata specifications
        https://packaging.python.org/specifications/core-metadata/

    - Required fields:
        - Metadata-Version
            * Version of the file format;
            * legal values are “1.0”, “1.1”, “1.2” and “2.1”.

        - Name
            * PEP 508 definition
            * The name of the distribution. The name field is the primary
        identifier for a distribution. A valid name consists only of ASCII
        letters and numbers, period, underscore and hyphen. It must start
        and end with a letter or number. Distribution names are limited to
        those which match the following regex (run with re.IGNORECASE):

        RE_PEP508_NAME = '^([A-Z0-9]|[A-Z0-9][A-Z0-9._-]*[A-Z0-9])$'

        - Version
            A string containing the distribution’s version number.
            This field must be in the format specified in PEP 440.

    - All other fields are optional:
        Platform - A Platform specification describing an operating system
            supported by the distribution which is not listed in the “Operating
            System” Trove classifiers

        Supported-Platform - Binary distributions containing a PKG-INFO file will
            use the Supported-Platform field in their metadata to specify the OS
            and CPU for which the binary distribution was compiled.

        Summary - A one-line summary of what the distribution does.

        Description - A longer description of the distribution.

        Description-Content-Type - A string stating the markup syntax (if any)
            used in the distribution’s description, so that tools can
            intelligently render the description. Historically, this was plain
            text or RST, but Markdown is common (RFC 7763)

            RFC 7763 - https://www.w3.org/Protocols/rfc1341/4_Content-Type.html

            Example:
            Description-Content-Type: text/markdown; charset=UTF-8; variant=GFM

            The type/subtype part has only a few legal values:
                text/plain
                text/x-rst
                text/markdown

            If a Description-Content-Type is not specified, then applications
            should attempt to render it as text/x-rst; charset=UTF-8 and fall
            back to text/plain if it is not valid rst.

        Keywords - A list of additional keywords, separated by commas, to be used
            to assist searching for the distribution in a larger catalog.

        Home-page - A string containing the URL for the distribution’s home page.

        Download-URL - A string containing the URL from which *this version* of
            the distribution can be downloaded.

        Author - A string containing the author’s name at a minimum;
            additional contact information may be provided.

        Author-email - A string containing the author’s e-mail address or a comma
            separated list of author e-mail addresses. It can contain a name and
            e-mail address in the legal forms for a RFC-822 From: header.

            e.g. "C. Schultz" <cschultz@example.com>

        Maintainer - A string containing the maintainer’s name at a minimum;
            additional contact information may be provided.

            Note that this field is intended for use when a project is being
            maintained by someone other than the original author: it should be
            omitted if it is identical to Author.

        Maintainer-email - A string containing the maintainer’s e-mail address
            with the same conditions as Author.

        License - Text indicating the license covering the distribution where the
            license is not a selection from the “License” Trove classifiers. See
            “Classifier” below. This field may also be used to specify a
            particular version of a license which is named via the Classifier
            field, or to indicate a variation or exception to such a license.

        Classifier - Each entry is a string giving a single classification value
            for the distribution. Classifiers are described in PEP 301, and the
            Python Package Index publishes a dynamic list of currently defined
            classifiers. This field may be followed by an environment marker after
            a semicolon.

            Curated list of classifiers: https://pypi.org/classifiers/

        Requires-Dist - Each entry contains a string naming some other distutils
            project required by this distribution.

        Requires-Python - This field specifies the Python version(s) that the
            distribution is guaranteed to be compatible with. Installation tools
            may look at this when picking which version of a project to install.

        Requires-External - Each entry contains a string describing some
            dependency in the system that the distribution is to be used. This
            field is intended to serve as a hint to downstream project maintainers,
            and has no semantics which are meaningful to the distutils
            distribution.

        Project-URL - A string containing a browsable URL for the project and a
            label for it, separated by a comma.

        Provides-Extra - A string containing the name of an optional feature. Must
            be a valid Python identifier. May be used to make a dependency
            conditional on whether the optional feature has been requested.

    Rarely Used:
        Provides-Dist - Each entry contains a string naming a Distutils project
            which is contained within this distribution.
        Obsoletes-Dist - Each entry contains a string describing a distutils
            project’s distribution which this distribution renders obsolete,
            meaning that the two projects should not be installed at the same time.

    - PEP 345 - Metadata for packages is deprecated
    - PEP 566 - Metadata for Python Software Packages 2.1 is used
    """

# 'Standard Library'
from os import linesep as NL

# 'third party'
from requests import RequestException, Response, get, request

from typing import List


def troves() -> (List[str]):
    """ Return updated list of Trove Classifiers from legacy endpoint.

        https://pypi.org/pypi?%3Aaction=list_classifiers
        """
    TROVES_API_URL: str = "https://pypi.org/pypi?%3Aaction=list_classifiers"
    response: requests.Response = requests.request("GET", TROVES_API_URL)
    if response.status_code == 200:
        return response.text.splitlines()
    else:
        return []


TROVES_LIST: List[str] = troves()
TROVES_STR: str = NL.join(TROVES_LIST)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
""" Part of the `AutoSys` package - utilities for macOS apps
        copyright (c) 2019 Michael Treanor
        https://www.github.com/skeptycal/autosys
        https://www.twitter.com/skeptycal

    `AutoSys` is licensed under the `MIT License
        `<https://opensource.org/licenses/MIT>`
    """

# 'package imports'
import cli.ansi

from cli.ansi import *

# __all__ = []



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
""" Part of the `AutoSys` package - utilities for macOS apps
        copyright (c) 2019 Michael Treanor
        https://www.github.com/skeptycal/autosys
        https://www.twitter.com/skeptycal

    `AutoSys` is licensed under the `MIT License
        `<https://opensource.org/licenses/MIT>`
    """

# 'Standard Library'
import functools
import inspect
import io
import os
import sys

# from functools import lru_cache, wraps
# from inspect import Parameter, signature
# from io import StringIO
# from os import fsync, path
from dataclasses import dataclass

# 'package imports'
import cli
import file_ops

from cli.terminal import hr

from typing import List

# from file_ops import *
# from file_ops.pytemp_dirs import *

_debug_: bool = True  # Turn on extra debug info display

_NAME: str = os.path.basename(__file__)
_HERE: str = os.path.dirname(__file__)


# !---------------------------------------------- Dynamic Class
# ref: https://stackoverflow.com/a/6581949/9878098


def choose_dyn_class(name):
    print(name)
    if name == "foo":

        class Foo(object):
            pass

        return Foo  # return the class, not an instance
    else:

        class Bar(object):
            pass

        return Bar


class Capturing(io.StringIO):
    """
        ### Creates a context manager to capture `stdout` text lines from code block.

        Returns:
            List[str] -- list containing text lines

        (This class only writes to disk file on exit. It is intended for use in short blocks of code as a temporary context manager. Long term usage increases the risk of data loss and will act as a memory leak.)

        ### Usage:

        ```
        with Capturing() as output:
            do_something(my_object)
        ```
        output is now a list containing the lines printed by the function call.

        ### Advanced usage:

        This can be done more than once and the results concatenated:

        ```
        with Capturing() as output:
            print('hello world')

        print('displays on screen')

        with Capturing(output) as output:  # note the constructor argument
            print('hello world2')

        print('done')
        print('output:', output)
        ```

        Output:

        ```
        displays on screen
        done
        output: ['hello world', 'hello world2']
        ```

        (Ref: https://stackoverflow.com/a/16571630)
        """

    def __init__(self):
        self._stdout = sys.stdout
        super().__init__()

    def __enter__(self):
        sys.stdout = self
        return super().__enter__()

    def __exit__(self, *args):
        """
            This function only writes to disk file on exit. It is intended for use in short blocks of code as a temporary context manager. Long term usage increases the risk of data loss and will act as a memory leak.
            """
        # self.extend(self._stringio.getvalue().splitlines())
        print(self.get_data(), file=sys.stderr)

        self.truncate(0)
        sys.stdout = self._stdout
        del self  # free up some memory

    def get_data(self, keepends: bool = False) -> List:
        return self.getvalue().splitlines(keepends=keepends)

    def write_to_disk(self):
        with TMP_HERE(prefix=_NAME) as f:
            f.writelines([_ for _ in self.get_data()])
            f.flush()
        os.fsync()  # commit disk writes


# wrapper to auto assign variables in __init__ method
# e.g.
# class Foo(object):

#     @auto_assign
#     def __init__(self, a, b, c=None, d=None, e=3):
#         pass


def auto_assign(func):
    # Signature:
    sig = signature(func)
    for name, param in sig.parameters.items():
        if param.kind in (Parameter.VAR_POSITIONAL, Parameter.VAR_KEYWORD):
            raise RuntimeError(
                "Unable to auto assign if *args or **kwargs in signature."
            )
    # Wrapper:

    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        for i, (name, param) in enumerate(sig.parameters.items()):
            # Skip 'self' param:
            if i == 0:
                continue
            # Search value in args, kwargs or defaults:
            if i - 1 < len(args):
                val = args[i - 1]
            elif name in kwargs:
                val = kwargs[name]
            else:
                val = param.default
            setattr(self, name, val)
        func(self, *args, **kwargs)

    return wrapper


def capture(capture_text):
    with Capturing() as output:
        if isinstance(capture_text, str):
            print(capture_text)
            return 0
        try:  # if iterable
            print(os.linesep.join([s for s in capture_text]))
            return 0
        except:
            pass
        try:
            print(str(capture_text))
            return 0
        except:
            return 1


def _test_():
    """
    #### Perform tests to produce additional debug info.

    (runs if _debug_ = True)
    """
    hr()
    print(f"{_NAME} is in DEBUG MODE.")

    print(choose_dyn_class)
    print(choose_dyn_class(auto_assign))
    hr()

    capture("capturing test - this should be captured to 'output'")
    # with Capturing() as output:
    #     print('capturing test - this should be captured to "output"')

    hr()
    print('capturing test - this should print to "stdout"')
    hr()
    print("\n --------------- done ...")


def _main_():
    """
    CLI script main entry point.
    """

    if _debug_:
        _test_()


if __name__ == "__main__":  # if script is loaded directly from CLI
    _main_()



#!/usr/bin/env python3
# -*- coding: UTF-8 -*-

# 'Standard Library'
import re

from collections import Counter

# 'third party'
import requests

SET_DEBUG: bool = True
RE_document_article = "^\b*"
RE_pattern_document_article: re.Pattern = re.compile(
    RE_document_article, re.IGNORECASE
)
FMT_URL_STATUSCODE_TEST = "url_test: GET request to {} returns {}"


def dbprint(*args, **kwargs):
    if SET_DEBUG:
        print(*args, **kwargs)


def safe_run(func):
    """ Decorator to handle errors in functions that often throw errors. Eliminates the need for multiple try/catch blocks.
    """

    def func_wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception as e:
            dbprint(e)
            return None

    return func_wrapper


def url_test(url, fmt=FMT_URL_STATUSCODE_TEST) -> str:
    """ Return formatted response from <url>."""
    r = requests.get(url)
    return fmt.format(url, r.status_code)


@safe_run
def get(url) -> (str, int):  # type: ignore
    """ Return text of response from <url>. """
    r = requests.get(url)
    if r.status_code != 200:
        return r.status_code
    else:
        return requests.get(url).text


def get_article(url) -> str:
    try:
        r: requests.Response = requests.get(url)
    except requests.RequestException as e:
        dbprint(f"{e=}")
        raise
    dbprint(f"{r.status_code=}")
    return r.text
    # if r.status_code
    # pass


def main():
    """
    CLI script main entry point.
    """
    url = "https://www.python.org"
    # dbprint(url_test(url))
    print(get("stuff"))

    med_list = [
        "https://medium.com/better-programming/stop-using-lists-for-everything-in-python-46fad15217f4",
        "https://medium.com/starts-with-a-bang/5-scientific-myths-you-probably-believe-about-the-universe-9a34597d7435",
    ]
    article_list = []
    for url in med_list:
        article_list.append(get_article(url))


if __name__ == "__main__":  # if script is loaded directly from CLI
    main()



# 'Standard Library'
from enum import IntEnum, unique


@unique
class ErrC(IntEnum):  # !------------------------ ErrC Class
    """ #### C-style error messages.
        Enum where members are also unique ints.

        - Reference: Advanced Bash-Scripting Guide
        <http://tldp.org/LDP/abs/html/exitcodes.html#EXITCODESREF>

        - `from /usr/include/sysexits.h`
        - Copyright (c) 1987, 1993
        - The Regents of the University of California.  All rights reserved.


        >The Linux Documentation Project has a list of reserved codes that also offers advice on what code to use for specific scenarios. These are the standard error codes in Linux or UNIX.

        GOALS:
        By creating this app, I wanted to answer these questions:

        - What are the historical roots of error codes in computing?
        - How can I use industry standard error codes in my programming?
        - How can I use an enum class in python in a useful way?
        """

    EX_OK = (0,)  # successful termination
    EX_ERROR = (1,)  # catchall for general errors
    EX_SHELLERR = (2,)  # misuse of shell builtins; missing keyword
    EX_USAGE = (64,)  # command line usage error
    EX_DATAERR = (65,)  # data format error
    EX_NOINPUT = (66,)  # cannot open input
    EX_NOUSER = (67,)  # addressee unknown
    EX_NOHOST = (68,)  # host name unknown
    EX_UNAVAILABL = (69,)  # service unavailable
    EX_SOFTWARE = (70,)  # internal software error
    EX_OSERR = (71,)  # system error (e.g., cant fork)
    EX_OSFILE = (72,)  # critical OS file missing
    EX_CANTCREAT = (73,)  # cant create (user) output file
    EX_IOERR = (74,)  # input/output error
    EX_TEMPFAIL = (75,)  # temp failure; user is invited to retry
    EX_PROTOCOL = (76,)  # remote error in protocol
    EX_NOPERM = (77,)  # permission denied
    EX_CONFIG = (78,)  # configuration error

    # Linux / Unix codes
    EX_CANTEXECUTE = (126,)  # command invoked cannot execute
    EX_NOTFOUND = (127,)  # command not found; possible $PATH error
    EX_BADARG = (128,)  # invalid argument
    EX_FATALARG = (129,)  # fatal error
    EX_CTRL_C = (130,)  # script terminated by Control-C


err = ErrC


class Error(Exception):
    """ Custom Exception handler.

        Generic class for C type error codes and messages
        """

    errno: int
    errmsg: str

    def __init__(self, *args, **kwargs):
        self.with_traceback = True
        self.args = args
        self.kwargs = kwargs
        # dbprint('error')
        # dbprint('args: ', self.args)
        # dbprint('kwargs: ', self.kwargs)
        for kwarg in self.kwargs:
            if kwarg == "errno":
                self.errno = self.kwargs[kwarg]
                self.errmsg = [
                    msg for num, msg in ErrC.items() if num == self.errno
                ]
                # dbprint(self.errno, ' ', self.errmsg)
            if kwarg == "message":
                self.message = self.kwargs[kwarg]
        Exception.__init__(self)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
""" gradient_dict.py """
# copyright (c) 2019 Michael Treanor
# https://www.github.com/skeptycal
# https://www.twitter.com/skeptycal


class Gradient_Dict(dict):
    """ A dictionary designed for efficient retrieval in large structures.
        Uses a gradient index allowing faster searching for keys. """

    pass



""" HashedIndex - InvertedIndex implementation using hash lists

    ```sh
    pip install hashedindex
    ```

    Fast and simple InvertedIndex implementation using hash lists (python dictionaries).

    Supports Python 3.5+

    Free software: BSD license

    Copyright (c) 2015, Michael Aquilina
    All rights reserved.

    Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.

    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.

    * Neither the name of hashedindex nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.

    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
    """

# 'Standard Library'
import math
import re
import unicodedata

from copy import copy
from string import ascii_letters, digits, punctuation

# 'package imports'
from hashed_index import *


# Stemmer interface which returns token unchanged
class NullStemmer:
    def stem(self, x):
        return x

    def __repr__(self):
        return "<NullStemmer>"


if True:
    _stopwords = frozenset()
    _accepted = frozenset(ascii_letters + digits + punctuation) - frozenset(
        "'"
    )

    _punctuation = copy(punctuation)
    _punctuation = _punctuation.replace("\\", "")
    _punctuation = _punctuation.replace("/", "")
    _punctuation = _punctuation.replace("-", "")

    _re_punctuation = re.compile("[%s]" % re.escape(_punctuation))
    _re_token = re.compile(r"[a-z0-9]+")

    _url_pattern = r"(https?:\/\/)?(([\da-z-]+)\.){1,2}.([a-z\.]{2,6})(/[\/\w \.-]*)*\/?(\?(\w+=\w+&?)+)?"
    _re_full_url = re.compile(r"^%s$" % _url_pattern)
    _re_url = re.compile(_url_pattern)


# Determining the best way to calculate tfidf is proving difficult,
# might need more advanced techniques
def tfidf(tf, df, corpus_size):
    """
    In information retrieval, tf–idf or TFIDF, short for term frequency–inverse document frequency, is a numerical statistic that is intended to reflect how important a word is to a document in a collection or corpus.

    It is often used as a weighting factor in searches of information retrieval, text mining, and user modeling. The tf–idf value increases proportionally to the number of times a word appears in the document and is offset by the number of documents in the corpus that contain the word, which helps to adjust for the fact that some words appear more frequently in general.

    tf–idf is one of the most popular term-weighting schemes today. A survey conducted in 2015 showed that 83% of text-based recommender systems in digital libraries use tf–idf.
    """
    if df and tf:
        return (1 + math.log(tf)) * math.log(corpus_size / df)
    else:
        return 0.0


def normalize_unicode(text):
    """
    Normalize any unicode characters to ascii equivalent

    Return the normal form form for the Unicode string unistr. Valid values for form are ‘NFC’, ‘NFKC’, ‘NFD’, and ‘NFKD’.

    https://docs.python.org/3.8/library/unicodedata.html#unicodedata.normalize
    """
    if isinstance(text, str):
        return (
            unicodedata.normalize("NFKD", text)
            .encode("ascii", "ignore")
            .decode("utf8")
        )
    else:
        return text


def get_ngrams(token_list, n=2):
    for i in range(len(token_list) - n + 1):
        yield token_list[i : i + n]


def word_tokenize(
    text, stopwords=_stopwords, ngrams=None, min_length=0, ignore_numeric=True
):
    """
    Parses the given text and yields tokens which represent words within
    the given text. Tokens are assumed to be divided by any form of
    whitespace character.
    """
    if ngrams is None:
        ngrams = 1

    text = re.sub(re.compile("'s"), "", text)  # Simple heuristic
    text = re.sub(_re_punctuation, "", text)

    matched_tokens = re.findall(_re_token, text.lower())
    for tokens in get_ngrams(matched_tokens, ngrams):
        for i in range(len(tokens)):
            tokens[i] = tokens[i].strip(punctuation)

            if len(tokens[i]) < min_length or tokens[i] in stopwords:
                break
            if ignore_numeric and isnumeric(tokens[i]):
                break
        else:
            yield tuple(tokens)


def isnumeric(text):
    """
    Returns a True if the text is purely numeric and False otherwise.
    """
    try:
        float(text)
    except ValueError:
        return False
    else:
        return True


def is_url(text):
    """
    Returns a True if the text is a url and False otherwise.
    """
    return bool(_re_full_url.match(text))



""" HashedIndex - InvertedIndex implementation using hash lists

    ```sh
    pip install hashedindex
    ```

    Fast and simple InvertedIndex implementation using hash lists (python dictionaries).

    Supports Python 3.5+

    Free software: BSD license

    Copyright (c) 2015, Michael Aquilina
    All rights reserved.

    Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.

    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.

    * Neither the name of hashedindex nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.

    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
    """

__author__ = "Michael Aquilina"
__email__ = "michaelaquilina@gmail.com"
__version__ = "0.5.0"

# 'Standard Library'
import collections
import functools
import math

DOCUMENT_DOES_NOT_EXIST = "The specified document does not exist"
TERM_DOES_NOT_EXIST = "The specified term does not exist"


class HashedIndex:
    """
    InvertedIndex structure in the form of a hash list implementation.
    """

    def __init__(self, initial_terms=None):
        """
        Construct a new HashedIndex. An optional list of initial terms
        may be passed which will be automatically added to the new HashedIndex.
        """
        self._documents = collections.Counter()
        self._terms = {}
        self._freeze = False
        if initial_terms is not None:
            for term in initial_terms:
                self._terms[term] = {}

    def __getitem__(self, term):
        return self._terms[term]

    def __contains__(self, term):
        return term in self._terms

    def __repr__(self):
        return "<HashedIndex: {} terms, {} documents>".format(
            len(self._terms), len(self._documents)
        )

    def __eq__(self, other):
        return (
            self._terms == other._terms and self._documents == other._documents
        )

    def clear(self):
        """
        Resets the HashedIndex to a clean state without any terms or documents.
        """
        self._terms = {}
        self._documents = collections.Counter()

    def freeze(self):
        """
        Freezes the HashedIndex, preventing any new terms from being added
        when calling add_term_occurrence.
        """
        self._freeze = True

    def unfreeze(self):
        """
        Unfreezes (thaws) the HashedIndex, allowing new terms to be added
        when calling add_term_occurrence.
        """
        self._freeze = False

    def add_term_occurrence(self, term, document):
        """
        Adds an occurrence of the term in the specified document.
        """
        if document not in self._documents:
            self._documents[document] = 0

        if term not in self._terms:
            if self._freeze:
                return
            else:
                self._terms[term] = collections.Counter()

        if document not in self._terms[term]:
            self._terms[term][document] = 0

        self._documents[document] += 1
        self._terms[term][document] += 1

    def get_total_term_frequency(self, term):
        """
        Gets the frequency of the specified term in the entire corpus
        added to the HashedIndex.
        """
        if term not in self._terms:
            raise IndexError(TERM_DOES_NOT_EXIST)

        return sum(self._terms[term].values())

    def get_term_frequency(self, term, document, normalized=False):
        """
        Returns the frequency of the term specified in the document.
        """
        if document not in self._documents:
            raise IndexError(DOCUMENT_DOES_NOT_EXIST)

        if term not in self._terms:
            raise IndexError(TERM_DOES_NOT_EXIST)

        result = self._terms[term].get(document, 0)
        if normalized:
            result /= self.get_document_length(document)

        return float(result)

    def get_document_frequency(self, term):
        """
        Returns the number of documents the specified term appears in.
        """
        if term not in self._terms:
            raise IndexError(TERM_DOES_NOT_EXIST)
        else:
            return len(self._terms[term])

    def get_document_length(self, document):
        """
        Returns the number of terms found within the specified document.
        """
        if document in self._documents:
            return self._documents[document]
        else:
            raise IndexError(DOCUMENT_DOES_NOT_EXIST)

    def get_documents(self, term):
        """
        Returns all documents related to the specified term in the
        form of a Counter object.
        """
        if term not in self._terms:
            raise IndexError(TERM_DOES_NOT_EXIST)
        else:
            return self._terms[term]

    def terms(self):
        return list(self._terms)

    def documents(self):
        return list(self._documents)

    def items(self):
        return self._terms

    def get_tfidf(self, term, document, normalized=False):
        """
        Returns the Term-Frequency Inverse-Document-Frequency value for the given
        term in the specified document. If normalized is True, term frequency will
        be divided by the document length.
        """
        tf = self.get_term_frequency(term, document)

        # Speeds up performance by avoiding extra calculations
        if tf != 0.0:
            # Add 1 to document frequency to prevent divide by 0
            # (Laplacian Correction)
            df = 1 + self.get_document_frequency(term)
            n = 2 + len(self._documents)

            if normalized:
                tf /= self.get_document_length(document)

            return tf * math.log10(n / df)
        else:
            return 0.0

    def get_total_tfidf(self, term, normalized=False):
        result = 0.0
        for document in self._documents:
            result += self.get_tfidf(term, document, normalized)
        return result

    def generate_document_vector(self, doc, mode="tfidf"):
        """
        Returns a representation of the specified document as a feature vector
        weighted according the mode specified (by default tf-dif).

        A custom weighting function can also be passed which receives the hashedindex
        instance, the selected term and document as parameters.

        The result will be returned in the form of a list. This can be converted
        into a numpy array if required using the `np.asarray` method
        Available built-in modes:
          * tfidf: Term Frequency Inverse Document Frequency
          * ntfidf: Normalized Term Frequency Inverse Document Frequency
          * tf: Term Frequency
          * ntf: Normalized Term Frequency
        """
        if mode == "tfidf":
            selected_function = HashedIndex.get_tfidf
        elif mode == "ntfidf":
            selected_function = functools.partial(
                HashedIndex.get_tfidf, normalized=True
            )
        elif mode == "tf":
            selected_function = HashedIndex.get_term_frequency
        elif mode == "ntf":
            selected_function = functools.partial(
                HashedIndex.get_term_frequency, normalized=True
            )
        elif hasattr(mode, "__call__"):
            selected_function = mode
        else:
            raise ValueError("Unexpected mode: %s", mode)

        result = []
        for term in self._terms:
            result.append(selected_function(self, term, doc))

        return result

    def generate_feature_matrix(self, mode="tfidf"):
        """
        Returns a feature matrix in the form of a list of lists which
        represents the terms and documents in this Inverted Index using
        the tf-idf weighting by default. The term counts in each
        document can alternatively be used by specifying scheme='count'.

        A custom weighting function can also be passed which receives a term
        and document as parameters.

        The size of the matrix is equal to m x n where m is
        the number of documents and n is the number of terms.

        The list-of-lists format returned by this function can be very easily
        converted to a numpy matrix if required using the `np.as_matrix`
        method.
        """
        result = []

        for doc in self._documents:
            result.append(self.generate_document_vector(doc, mode))

        return result

    def prune(self, min_value=None, max_value=None, use_percentile=False):
        n_documents = len(self._documents)

        garbage = []
        for term in self.terms():
            freq = self.get_document_frequency(term)
            if use_percentile:
                freq /= n_documents

            if min_value is not None and freq < min_value:
                garbage.append(term)

            if max_value is not None and freq > max_value:
                garbage.append(term)

        for term in garbage:
            del self._terms[term]

    def to_dict(self):
        return {
            "documents": self._documents,
            "terms": self._terms,
        }

    def from_dict(self, data):
        self._documents = collections.Counter(data["documents"])
        self._terms = {}
        for term in data["terms"]:
            self._terms[term] = collections.Counter(data["terms"][term])


def merge(index_list):
    result = HashedIndex()

    for index in index_list:
        first_index = result
        second_index = index

        assert isinstance(second_index, HashedIndex)

        for term in second_index.terms():
            if term in first_index._terms and term in second_index._terms:
                result._terms[term] = (
                    first_index._terms[term] + second_index._terms[term]
                )
            elif term in second_index._terms:
                result._terms[term] = second_index._terms[term]
            else:  # pragma: nocover
                raise ValueError(
                    "I dont know how the hell you managed to get here"
                )

        result._documents = first_index._documents + second_index._documents

    return result



if True:  # !------------------------ config
    import logging
    from pathlib import Path
    from typing import List, Dict
    from autosys.cli import terminal

    _debug_: bool = True  # True => use Debug features
    _verbose_: int = 0  # Verbosity level => 0 - 4
    _log_flag_: bool = _debug_ and True  # True => log to file

    DEBUG_COLOR: str = "\x1B[38;5;178m"  # private ansi CLI color code
    RESET: str = "\x1B[0m"  # private ansi CLI reset code


if True:  # !------------------------ CLI display utilities

    def db_column_ruler(n: int = 5, cols: int = 79):
        """ Print a column ruler of width 'col'. """
        # dbprint('column ruler col: ', col)
        col = cols // n
        dbprint(f'__width:{col:>d}{s80("_", (col - 1) * 10 - 1)}', sep="")
        dbprint("".join([f"         {i}" for i in range(1, col)]))
        dbprint("")
        dbprint(f"1234567890" * (col))
        dbprint(s80("=", col * 10))


# if True:  # !------------------------ Debug Utilities

#     def v_name(**var):
#         """ #### Return the name of a keyword variable as a string.

#             Example:
#             ```
#                 the_value = 'some_value'
#                 print(v_name(the_value=the_value))
#             ```

#             This method is 2x faster than f_name (n = 10000)
#             - 'profile_v_name'  4.98 ms
#             - 'profile_f_name'  10.97 ms

#             >(deprecated: f_name()  :  Return the name of a keyword variable as a string. An alternative to using a list comprehension with function <v_name> that manipulates f-strings directly.)
#             """
#         return [_ for _ in var][0]

#     def tryit(func):
#         """ Decorator to wrap function in a try/except block.

#             Will only catch the first error ... best used for short, quick functions...

#             """
#         def tried(*args, **kwargs):
#             try:
#                 result = func(*args, **kwargs)
#             except Exception as e:
#                 log_it(e)
#                 result = "failed..."
#             else:
#                 # print(args)  # TODO <-- your code here
#                 result = "tried..."
#             finally:
#                 dbprint(result)
#             return result

#         return tried

#     def get_class_name(obj, verbose=True):
#         _cls_name: str = obj.__class__.__name__
#         if not verbose:
#             return _cls_name
#         return f"{_cls_name=}"

# if True:  # !------------------------ Logging Utilities

#     def info(*args):
#         logging.info(*args)

#     def log_error(*args):
#         """ Error reporting in debug mode.

#             current (temporary) behavior: print error messages to <stderr> using dbprint. (if _debug_ flag is True)
#             """
#         # TODO - fix temporary functionality
#         if _log_flag_:
#             logging.info(*args)
#             dbprint(*args)  # ! temp

#     def logex(func):
#         """ Decorator to catch and log errors. """
#         if _log_flag_:
#             try:
#                 result = func()
#             except IOError as e:
#                 log_error(
#                     f"logex caused an error while reporting <{func}>: {e}.")
#             else:
#                 if isinstance(result, Exception):
#                     log_error(f"Function <{func}> caught an error: {e}.")
#             finally:
#                 del e

#     def verbose(v: int, *args, **kwargs):
#         """ Print based on '_verbose_' allowed verbosity level.

#             v: int - requested verbosity level
#             """
#         try:  # if _verbose_ constant does not exist, skip this function
#             _verbose_ == 0
#         except NameError:
#             return -1
#         if _verbose_ >= v:
#             if v < 2:
#                 kwargs["file"] = "sys.stdout"
#                 print(*args, **kwargs)
#             elif v == 2:
#                 kwargs["file"] = "sys.stderr"
#                 print(*args, **kwargs)



# -*- coding: utf-8 -*-


# Python program to print
# colored text and background
def print_format_table():
    """
    prints table of formatted text format options
    """
    for style in range(8):
        for fg in range(30, 38):
            s1 = ""
            for bg in range(40, 48):
                format = ";".join([str(style), str(fg), str(bg)])
                s1 += "\x1b[%sm %s \x1b[0m" % (format, format)
            print(s1)
        print("\n")


print_format_table()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
""" AutoSys Setup
    ---
    Part of the [AutoSys][1] package

    Copyright (c) 2018 [Michael Treanor][2]

    AutoSys is licensed under the [MIT License][3]

    [1]: https://www.github.com/skeptycal/autosys
    [2]: https://www.twitter.com/skeptycal
    [3]: https://opensource.org/licenses/MIT
    """

# Note: To use the 'upload' functionality of this file, you must:
#   $ pip install twine

import logging
import os

from os import linesep as NL
from pathlib import Path
from sys import argv as _argv, path as PYTHONPATH

from setuptools import find_namespace_packages, setup

from typing import Dict, Final, List, Optional, Sequence, Tuple

# ? #################################### config

try:
    DEFAULT_ENCODING
except:
    from locale import getpreferredencoding
    DEFAULT_ENCODING: str = getpreferredencoding(do_setlocale=True) or "utf-8"

#! DEBUG - run some live tests ... set 'False' for production !!!
_debug_: bool = False


here = Path(__file__).resolve().parent

if here not in PYTHONPATH:
    PYTHONPATH.insert(0, here)

logger = logging.getLogger(__name__)


# ? #################################### packaging utilities.
def table_print(data: (Dict, Sequence), **kwargs):
    tmp: List = []
    if isinstance(data, dict):
        tmp.extend(
            [f"{str(k):<15.15} :  {repr(v):<45.45}" for k, v in data.items()])
    elif isinstance(data, (list, tuple, set)):
        for x in data:
            try:
                tmp.append(f"{str(x):<15.15} :  {repr(f'{x}'):<45.45}")
            except:
                tmp.append(f"{str(x)}")
    else:
        raise TypeError('Parameter must be an iterable Mapping or Sequence.')
    print(NL.join(tmp), **kwargs)


def pip_safe_name(s: str):
    """ Return a name that is converted to pypi safe format. """
    return s.lower().replace("-", "_").replace(" ", "_")


def readme(file_name: str = "readme.md"):
    """ Returns the text of the README file

        The default file is `README.md` and is *NOT* case sensitive. (e.g. `README` is the same as `readme`)
        Can load *any* text file, but the default search path is setup for readme files

        ```
        Search path = ["readme.md", "readme.rst", "readme", "readme.txt"]
        ```

        Example:

        ```
        long_description=readme()
        ```
        """

    search_list = ["readme.md", "readme.rst", "readme", "readme.txt"]
    if file_name not in search_list:
        search_list.insert(0, file_name)
    found: bool = False
    for searchfile in search_list:
        for parent in Path(__file__).resolve().parents:
            find_path = Path(parent / searchfile)
            if find_path.exists():
                found = True
                # print(find_path)
                break
        if found:
            break
    if found:
        try:
            with open(find_path, mode="r", encoding=DEFAULT_ENCODING) as f:
                return f.read()
        except IOError as e:
            raise IOError(f"Cannot read from the 'readme' file '{find_path}'")
    else:
        raise FileNotFoundError(
            f"Cannot find project 'readme' file in project tree. Search list = {search_list}"
        )


# ? #################################### package metadata.
__version__: str = '0.0.1'


NAME: str = pip_safe_name("PathPack")

VERSION: str = __version__  # "0.4.4"
VERSION_INFO: Tuple[int] = VERSION.split(".")
DESCRIPTION: str = "A subclass of pathlib and allows subclassing directly from Path, etc."
REQUIRES_PYTHON: str = ">=3.8.0"
PACKAGE_DIR: Dict = {'': f'{NAME}'}
PACKAGE_EXCLUDE: List[str] = ['*test*', '*bak*']
LICENSE: str = "MIT"
LONG_DESCRIPTION: str = readme()
LONG_DESCRIPTION_CONTENT_TYPE: str = "text/markdown"
# LONG_DESCRIPTION_CONTENT_TYPE="text/x-rst",
AUTHOR: str = "Michael Treanor"
AUTHOR_EMAIL: str = "skeptycal@gmail.com"
MAINTAINER: str = ""
MAINTAINER_EMAIL: str = ""
URL: str = f"https://skeptycal.github.io/{NAME}/"
DOWNLOAD_URL: str = f"https://github.com/skeptycal/{NAME}/archive/{VERSION}.tar.gz"
ZIP_SAFE: bool = False
INCLUDE_PACKAGE_DATA: bool = True
# What packages are required for this module to be executed?
REQUIRED: List[str] = [
    "aiocontextvars>=0.2.0 ; python_version<'3.7'",
    "colorama>=0.3.4 ; sys_platform=='win32'",
    "win32-setctime>=1.0.0 ; sys_platform=='win32'",
]

# What packages are optional?
EXTRAS: Dict = {
    ":python_version < '3.5'": ["typing==3.6.1", ],
    "dev": [
        "black>=19.3b0 ; python_version>='3.8'",
        "codecov>=2.0.15",
        "flake8>=3.7.7",
        "isort>=4.3.20",
        "tox>=3.9.0",
        "tox-travis>=0.12",
        "pytest>=4.6.2",
        "pytest-cov>=2.7.1",
        "Sphinx>=2.2.1",
        "sphinx-autobuild>=0.7.1",
        "sphinx-rtd-theme>=0.4.3",
    ]
}

PACKAGE_DATA: Dict = {
    # If any package contains these files, include them:
    "": [
        "*.txt",
        "*.rst",
        "*.md",
        "*.ini",
        "*.png",
        "*.jpg",
        "*.py",
        "__init__.pyi",
        "py.typed",
    ]
}

PROJECT_URLS: Dict = {
    "Website": f"https://skeptycal.github.io/{NAME}/",
    "Documentation": f"https://skeptycal.github.io/{NAME}/docs",
    "Source Code": f"https://www.github.com/skeptycal/{NAME}/",
    "Changelog":
    f"https://github.com/skeptycal/{NAME}/blob/master/CHANGELOG.md",
}

KEYWORDS: List = [
    "application",
    "macOS",
    "dev",
    "devops",
    "cache",
    "utilities",
    "cli",
    "python",
    "cython",
    "text",
    "console",
    "log",
    "debug",
    "test",
    "testing",
    "logging",
    "logger",
]
CLASSIFIERS: List = [
    "Development Status :: 4 - Beta",
    "License :: OSI Approved :: MIT License",
    "Environment :: Console",
    "Environment :: MacOS X",
    "Intended Audience :: Developers",
    "Natural Language :: English",
    "Operating System :: MacOS",
    "Operating System :: OS Independent",
    "Programming Language :: Cython",
    "Programming Language :: Python",
    "Programming Language :: Python :: 3 :: Only",
    "Programming Language :: Python :: 3",
    # These are the Python versions tested; it may work on others
    "Programming Language :: Python :: 3.8",
    "Programming Language :: Python :: 3.9",
    "Programming Language :: Python :: Implementation :: CPython",
    "Programming Language :: Python :: Implementation :: PyPy",
    "Topic :: Software Development :: Libraries :: Python Modules",
    "Topic :: Software Development :: Testing",
    "Topic :: Utilities",
]


def main(args=_argv[1:], ):  # ? ############################## Setup!
    global _debug_
    if 'debug' in args:
        _debug_ = True
    if _debug_:  # do some live tests if setup process has changed ...
        print(f"{NAME=}")
        print(f"{VERSION=}")
        # table_print(_vars)
    else:  # run setup ...
        setup(
            name=NAME,
            version=VERSION,
            description=DESCRIPTION,
            python_requires=REQUIRES_PYTHON,
            package_dir=PACKAGE_DIR,
            packages=find_namespace_packages(
                f'{NAME}', exclude=PACKAGE_EXCLUDE),
            # py_modules=[f"{NAME}"],
            license=LICENSE,
            long_description=LONG_DESCRIPTION,
            long_description_content_type=LONG_DESCRIPTION_CONTENT_TYPE,
            author=AUTHOR,
            author_email=AUTHOR_EMAIL,
            maintainer=MAINTAINER or AUTHOR,
            maintainer_email=MAINTAINER_EMAIL or AUTHOR_EMAIL,
            url=URL,
            download_url=DOWNLOAD_URL,
            zip_safe=ZIP_SAFE,
            include_package_data=INCLUDE_PACKAGE_DATA,
            # setup_requires=["isort"],
            install_requires=REQUIRED,
            extras_require=EXTRAS,
            package_data=PACKAGE_DATA,
            project_urls=PROJECT_URLS,
            keywords=KEYWORDS,
            classifiers=CLASSIFIERS,
        )


if __name__ == "__main__":
    main()

else:
    pass
    '''
    # references ...
    colorama_Fore: Dict[str, str] = {
        'BLACK': '\x1b[30m',
        'BLUE': '\x1b[34m',
        'CYAN': '\x1b[36m',
        'GREEN': '\x1b[32m',
        'LIGHTBLACK_EX': '\x1b[90m',
        'LIGHTBLUE_EX': '\x1b[94m',
        'LIGHTCYAN_EX': '\x1b[96m',
        'LIGHTGREEN_EX': '\x1b[92m',
        'LIGHTMAGENTA_EX': '\x1b[95m',
        'LIGHTRED_EX': '\x1b[91m',
        'LIGHTWHITE_EX': '\x1b[97m',
        'LIGHTYELLOW_EX': '\x1b[93m',
        'MAGENTA': '\x1b[35m',
        'RED': '\x1b[31m',
        'RESET': '\x1b[39m',
        'WHITE': '\x1b[37m',
        'YELLOW': '\x1b[33m',
    }
    colorama_Back: Dict[str, str] = {
        'BLACK': '\x1b[40m',
        'BLUE': '\x1b[44m',
        'CYAN': '\x1b[46m',
        'GREEN': '\x1b[42m',
        'LIGHTBLACK_EX': '\x1b[100m',
        'LIGHTBLUE_EX': '\x1b[104m',
        'LIGHTCYAN_EX': '\x1b[106m',
        'LIGHTGREEN_EX': '\x1b[102m',
        'LIGHTMAGENTA_EX': '\x1b[105m',
        'LIGHTRED_EX': '\x1b[101m',
        'LIGHTWHITE_EX': '\x1b[107m',
        'LIGHTYELLOW_EX': '\x1b[103m',
        'MAGENTA': '\x1b[45m',
        'RED': '\x1b[41m',
        'RESET': '\x1b[49m',
        'WHITE': '\x1b[47m',
        'YELLOW': '\x1b[43m'
    }
    colorama_Style: Dict[str, str] = {
        'BRIGHT': '\x1b[1m',
        'DIM': '\x1b[2m',
        'NORMAL': '\x1b[22m',
        'RESET_ALL': '\x1b[0m'
    }
    '''



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
""" Terminal
    ---
    terminal - Utilities for macOS terminal io.

    AutoSys
    ---
    Part of the [AutoSys][1] package

    Copyright (c) 2018 [Michael Treanor][2]

    AutoSys is licensed under the [MIT License][3]

    [1]: https://www.github.com/skeptycal/autosys
    [2]: https://www.twitter.com/skeptycal
    [3]: https://opensource.org/licenses/MIT
    """

# from dataclasses import Field, dataclass, field
# from io import TextIOWrapper
# 'package imports'
# from os import environ as ENV, linesep as NL
# from platform import platform
from autosys.cli import *
from platform import platform

__all__ = [
    "BasicColors",
    "br",
    "CR",
    "DEFAULT_COLOR",
    "hr",
    "NL",
    "PLATFORM",
    "rprint",
    "s80",
    "SUPPORTS_COLOR",
    "Terminal",
    "vprint",
]

PLATFORM = platform()
DEFAULT_COLOR = "MAIN"
CR: str = "\r"

if True:  # !------------------------ CLI display utilities

    def hr(s: str = "-", width: int = 50, print_it: bool = True):
        """ yes, a dashed line inspired by `<HR />`
            ---
            s - character to use for line

            width - duh ... the width
            """
        if not print_it:
            return s * width
        else:
            print(f"{s}" * width)

    def br(n: int = 1):
        """ yes, a newline inspired by `<BR />`
            ---
            n - number of line breaks
            """
        print(NL * n, end='')

    def s80(s: str = "=", n: int = 1):
        """ string80
            ---
            print a 79 character repeating string

            n - number of times to print it
            """
        for _ in range(n):
            print(s * 79)

    def vprint(var_name: str, print_it: bool = True):
        fmt = f"{var_name}"
        if print_it:
            print(fmt)
        else:
            return fmt

    def rprint(*args, **kwargs):
        """ Carriage return without a newline.
            ---
            Moves back to the start of the line.

            Cool for progress bars and counters ...
            """
        try:
            kwargs['end'] = ''
            kwargs['sep'] = ''
        except:
            pass
        print(CR, *args, **kwargs)


@dataclass
class Terminal:  # !------------------------ Terminal Class
    _SUPPORTS_COLOR: bool = False
    _SIZE: Tuple[int, int] = None
    _stream: TextIOWrapper = stdout
    DEFAULT_TERMINAL_SIZE: NamedTuple = (80, 24)

    # !------------------------------ properties

    @property
    def SUPPORTS_COLOR(self) -> (bool):
        if not self._SUPPORTS_COLOR:
            self._SUPPORTS_COLOR = self._get_supports_color()
        return self._SUPPORTS_COLOR

    @property
    def SIZE(self) -> (Tuple[int, int]):
        if not self._SIZE:
            self._SIZE = self._get_terminal_size()
        return self._SIZE

    @property
    def cols(self) -> (int):
        return self.SIZE[0]

    @property
    def rows(self) -> (int):
        return self.SIZE[1]

    # !------------------------------ methods

    def __str__(self) -> (str):
        return f"Terminal object (Supports color? {self.SUPPORTS_COLOR})"

    def __repr__(self) -> (str):
        _repr_list = [self.__str__()]
        # ? s += f'Terminal properties:{NL}'
        for p in sorted(dir()):
            if not p.startswith("_"):
                f = f"{p}"
                _repr_list.append(f"  {p}: {eval(f)}")
        return NL.join(_repr_list)

    def _ioctl_get_win_size(self, fd: int) -> (Tuple[int, int], Exception):
        """ Return Tuple 'cr' - columns, rows of tty """
        try:
            from fcntl import ioctl
            from struct import unpack
            from termios import TIOCGWINSZ

            retval = unpack("hh", ioctl(fd, TIOCGWINSZ, "1234"))
            log.info(f"_ioctl_get_win_size returns: {retval=}")
            return int(retval[1]), int(retval[0])
        except Exception as e:
            return e

    def _get_terminal_size(self) -> (Tuple[int, int]):
        """ Return terminal SIZE as a tuple(COLS, ROWS).

                Attempts to locate a valid terminal SIZE using various fallback methods.

                Default is 80 columns x 24 rows.

                Reference: https://stackoverflow.com/questions/566746/how-to-get-linux-console-window-width-in-python
                """
        # import os
        import sys

        cr: Tuple[int, int] = None

        try:
            print(1 / 0)
            cr = (self._ioctl_get_win_size(0) or self._ioctl_get_win_size(1)
                  or self._ioctl_get_win_size(2))
            if cr:
                log.info(f"cr from 'or' ioctl's: {cr}")
                return cr
        except Exception as e:
            log.error(e)
        log.info(f"no cr from 'or' ioctl's")

        try:
            from os import close as _close, open as _open, ctermid, O_RDONLY

            fd = _open(ctermid(), O_RDONLY)
            cr = self._ioctl_get_win_size(fd)
            _close(fd)
            with _open(ctermid(), O_RDONLY) as fd:
                cr = self._ioctl_get_win_size(fd)
                log.info(f"cr from cterm(): {cr}")
            if cr:
                log.info(f"cr from cterm(): {cr}")
                return cr
        except Exception as e:
            log.error(e)
        log.info(f"no cr from cterm()")

        try:
            if ENV.get("LINES") and ENV.get("COLUMNS"):
                cr = (ENV.get("LINES"), ENV.get("COLUMNS"))
            if cr:
                log.info(f"env.get returns {cr}")
                return cr
        except Exception as e:
            log.error(e)
        log.info(f"no return from env.get")

        try:
            # 'fallback' also sets default if nothing else has worked
            from shutil import get_terminal_size as _SH_SIZE

            cr = _SH_SIZE(fallback=Terminal.DEFAULT_TERMINAL_SIZE)
            if cr:
                log.info(
                    f"shutil.get_terminal_size returns ({cr.columns},{cr.lines})"
                )
                return (cr.columns, cr.lines)
        except Exception as e:
            log.error(e)
        log.info("no return from shutil.get_terminal_size")
        log.info(f"using {Terminal.DEFAULT_TERMINAL_SIZE} for cr")

        return (
            Terminal.DEFAULT_TERMINAL_SIZE[0],
            Terminal.DEFAULT_TERMINAL_SIZE[1],
        )

    def _get_supports_color(self) -> (bool):
        """ generic script level stderr output characteristics """
        self._IS_A_TTY: bool = self._stream.isatty() and hasattr(
            self._stream, "isatty")
        # if self._IS_A_TTY:
        #     return True
        self._IS_PPC: bool = PLATFORM == "Pocket PC"
        self._IS_WIN32: bool = PLATFORM == "win32"
        self._IS_ANSICON: bool = "ANSICON" in ENV
        self._IS_WIN_COLOR: bool = self._IS_WIN32 and self._IS_ANSICON
        self._IS_EDGE_CASE: bool = self._IS_WIN_COLOR or self._IS_PPC
        return self._IS_A_TTY or self._IS_EDGE_CASE

    def _show_debug_info(self) -> (None):
        hr(s="=")
        print("Terminal Properties:")
        hr()
        constants = {k: eval(f"term.{k}") for k in dir(term) if k.isupper()}
        for k, v in constants.items():
            print(k, v)
        print("---------------------------------------")

    def out(self, *args, sep=" ", end=NL, flush=False) -> (bool, Exception):
        """ send output to the stream. catch and return errors. """
        try:
            print(*args, sep=sep, end=end, flush=flush, file=self._stream)
            return False
        except Exception as e:
            return e


term = Terminal()

SUPPORTS_COLOR: bool = term.SUPPORTS_COLOR


# some basic colors ..
class BasicColors:
    MAIN: str = "\x1B[38;5;229m" * SUPPORTS_COLOR
    WARN: str = "\x1B[38;5;203m" * SUPPORTS_COLOR
    BLUE: str = "\x1B[38;5;38m" * SUPPORTS_COLOR
    GO: str = "\x1B[38;5;28m" * SUPPORTS_COLOR
    CHERRY: str = "\x1B[38;5;124m" * SUPPORTS_COLOR
    CANARY: str = "\x1B[38;5;226m" * SUPPORTS_COLOR
    ATTN: str = "\x1B[38;5;178m" * SUPPORTS_COLOR
    RAIN: str = "\x1B[38;5;93m" * SUPPORTS_COLOR
    WHITE: str = "\x1B[37m" * SUPPORTS_COLOR
    RESET: str = "\x1B[0m" * SUPPORTS_COLOR


if True:

    def _test_terminal_():
        from pprint import pprint
        hr()
        log.var("PLATFORM")
        log.var("term")
        log.var("term.SUPPORTS_COLOR")
        log.var("SUPPORTS_COLOR")
        log.var("term._stream")
        log.var("term._SIZE")
        log.var("term.SIZE")
        log.info(f"Terminal SIZE is set to ({term.cols}, {term.rows})")
        hr()
        print(str(lc))
        print(lc.str())
        print(dir(lc))

    # _test_terminal_()

    # term._show_debug_info()

    def _slow_progress_example():
        from time import sleep

        hr()
        br()
        hr()
        s80('*')
        hr()
        for i in range(50):
            rprint('=' * i, '>', ' ' * (80 - i))
            sleep(0.02)
        for i in range(50, 0, -1):
            rprint('=' * i, '>', ' ' * (80 - i))
            sleep(0.02)

    # _slow_progress_example()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# 'Standard Library'
from dataclasses import dataclass
from enum import Enum, IntEnum, unique
from time import sleep

# 'package imports'
from autosys import *
from cli.supports_color import SUPPORTS_COLOR

# 'third party'
import numpy as np

from typing import Dict, Iterable

NAN = object()
NUL = ""

'''
@unique
class NpAlign(Enum):
    left = "<"
    right = ">"
    center = "^"


@unique
class FB(Enum):
    fg = 3
    bg = 4


class Code8:
    black: str = "0"
    red: str = "1"
    green: str = "2"
    yellow: str = "3"
    blue: str = "4"
    magenta: str = "5"
    cyan: str = "6"
    white: str = "7"


class Code16:
    black: str = "0"
    red: str = "1"
    green: str = "2"
    yellow: str = "3"
    blue: str = "4"
    magenta: str = "5"
    cyan: str = "6"
    white: str = "7"
    bright_black: str = "0;1"
    bright_red: str = "1;1"
    bright_green: str = "2;1"
    bright_yellow: str = "3;1"
    bright_blue: str = "4;1"
    bright_magenta: str = "5;1"
    bright_cyan: str = "6;1"
    bright_white: str = "7;1"


class Effects:
    bold: str = "\u001b[1m"
    underline: str = "\u001b[4m"
    reverse: str = "\u001b[7m"
    reset: str = "\u001b[0m"


class Ansi(str):
    reset: str = "\u001b[0m"

    def __init__(self, *args, **kwargs):
        super(*args, **kwargs)

    def _color8(self, c: Code8 = "1", fb: FB = FB.fg):
        """ get either fg or bg color (8) - default fg """
        return f"{self}\u001b[{fb}{c}m"

    def _color16(self, c: Code16 = Code16.blue, fb: FB = FB.fg):
        """ get either fg or bg color (16) - default fg """
        return f"{self}\u001b[{fb}{c}m"

    def _color256(self, c="229", fb: FB = FB.fg):
        """ get either fg or bg color (256) - default fg """
        return f"\u001b[{fb}8;5;{c}m"

    def fb(self, fg="0", bg="229"):
        """ <F>oreground / <B>ackground 256 color combo """
        return f"\u001b[38;5;{fg}m\u001b[48;5;{bg}m"


class ColorToggle:
    a = Ansi()
    DEFAULT_NORMAL = a.fb(fg=2, bg=229)
    DEFAULT_REVERSE = a.fb(fg=16, bg=229)

    reset: str = "\u001b[0m"

    def __init__(
        self,
        norm=DEFAULT_NORMAL,
        rev=DEFAULT_REVERSE,
        rows=True,
        cols=True,
        checkerboard=True,
    ):
        # set colors
        self._normal = norm
        self._reverse = rev

        # set features
        self._on: bool = True  # turn all colors on and off
        self._rows: bool = rows  # color rows
        self._cols: bool = cols  # color columns
        self._checkerboard: bool = checkerboard  # do checkerboard?

        # store initial state
        self._row_state: bool = self._rows  # state of rows
        self._col_state: bool = self._cols  # state of columns

    @property
    def row(self):
        self._row_state = not self._row_state
        # if self._checkerboard:
        #     return self.col
        return self

    @property
    def col(self):
        self._col_state = not self._col_state
        return self

    def __str__(self):
        return self._row_state and self._normal or self._reverse


class NpChart:
    def __init__(
        self,
        a: Iterable,
        size: int = 4,
        padding: int = 1,
        pad: str = " ",
        align: NpAlign = NpAlign.left,
    ):

        pass


def np_row(
    a: Iterable,
    size: int = 4,
    padding: int = 1,
    pad: str = " ",
    color: bool = True,
    alt_rows: bool = True,
    alt_columns: bool = True,
    checkerboard: bool = True,
    align: NpAlign = NpAlign.left,
):
    """ Format a row of data for cli output.

        param       description                             default
        ===========================================================
        a -         any iterable data set
        size -      space allowed for each data point       4
        padding -   space between each data point           1
        pad -       character used for padding              ' '
        alt_color - should it alternate colors?             False
        align -     alignment of data in cells              '<'

        """

    ct = ColorToggle() if SUPPORTS_COLOR else ""
    return f'{ct.row}{"".join([f"{ct.col}{x:{align}{size}}{pad*padding}" for x in a])}{ct.reset}'


a = np.arange(5)
arithmetic_lst = ["a", "a + 10", "a - 10", "a * 2", "a / 2", "a // 2"]
for arithmetic in arithmetic_lst:
    evl = eval(arithmetic)
    fmt = np_row(evl, 3, align=NpAlign.right)
    print(f"{arithmetic:>10}: {fmt}")
print()

'''
# print methods of numpy.ndarray without dunders
# print([x for x in dir(evl) if not x.startswith("_")])



#!/usr/bin/env python3
# -*- coding: UTF-8 -*-
"""
Usage:
    err [OPTION]... N

Options:
    -n, --quiet, --silent     suppress most output
    -d, --debug               annotate program execution
    -o file, --output=file    send output to this file
    -l N, --line-length=N     specify the desired line-wrap length
    -z, --null-data           separate lines by NUL characters
    --help                    display this help and exit
    --version                 output version information and exit



err.py - Provides processing, formating, output, and logging for errors.
    Imported module provides the Err class that is used to catch and process errors in python programs.

The CLI version provides lookup information on errors and features for testing/development,
  but is not very functional as an ongoing solution.

AutoSys home page: <https://skeptycal.github.io/autosys/>.
General help using GNU software: <https://www.gnu.org/gethelp/>.
Bug reports and collaboration: <https://github.com/skeptycal/autosys>.

Usage: err [-dhtv] n


    DEFAULT_ARGS: Dict[str, List[Sequence[str]]] = {
        "debug": [
            'Provide debug feedback - not for production use.',
            ['debug', '-d', '--debug'],
        ],
        "help": [
            'Display command information.',
            ['help', '-h', '--help'],
        ],
        "test": [
            'Run test scripts.',
            ['test', '-t', '--test'],
        ],
        "version": [
            'Display version information.',
            ['version', '-v', '--version'],
        ],

"""

# 'Standard Library'
import enum

from enum import Enum, auto

# 'package imports'
import sentry_sdk

from typing import Dict, List, Sequence, Tuple

sentry_sdk.init("https://853c4343706c4a139a7f1c43b7eb2a51@sentry.io/1887434")
sentry_sdk.init(release="as-anansi@0.3.0")

# ! ############# Sentry Test - do not use
# ! division_by_zero = 1 / 0


class CLI:
    """ Command line interface with argument parsing, default values, and """

    from sys import argv
    from docopt import docopt

    # defaults setup from common CLI interfaces
    DEFAULT_ARGS: Dict[str, List[Sequence[str]]] = {
        "debug": [
            "Provide debug feedback - not for production use.",
            ["debug", "-d", "--debug"],
        ],
        "help": ["Display command information.", ["help", "-h", "--help"],],
        "test": ["Run test scripts.", ["test", "-t", "--test"],],
        "version": [
            "Display version information.",
            ["version", "-v", "--version"],
        ],
    }

    # FLAGS_ARGS are autofilled with keys from DEFAULT_ARGS
    FLAGS_ARGS = [f"FL_{k.upper()}" for k in DEFAULT_ARGS.keys()]

    def __init__(self, args):
        super().__init__()
        args = docopt(__doc__, version=f"AutoSys {__version__}")

        ARGS = self.DEFAULT_ARGS

    @classmethod
    def is_arg(cls, arg, test: str) -> bool:
        # if arg[0:1] == '--':
        #     arg = arg[2:]
        args = cls.DEFAULT_ARGS.get(test)
        s = f'Is "{arg}" in {args}? ... {arg in args}'  # type:ignore
        print(s)
        return arg in args  # type:ignore

    @classmethod
    def _add_defaults(cls, d):
        try:
            cls.DEFAULT_ARGS = cls.DEFAULT_ARGS.update(d)
        except:  # noqa
            pass

    @classmethod
    def _load_new_defaults(cls, d):
        try:
            cls.DEFAULT_ARGS = d
        except:  # noqa
            pass


class Err(Enum):
    """ C++ style error messages """

    # setup C++ style error messages
    # reference: Advanced Bash-Scripting Guide
    #   <http://tldp.org/LDP/abs/html/exitcodes.html#EXITCODESREF>

    # from /usr/include/sysexits.h
    # Copyright (c) 1987, 1993
    # The Regents of the University of California.  All rights reserved.
    ERR_OK_ = 0  # successful termination

    # - The Linux Documentation Project has a list of reserved codes that
    #   also offers advice on what code to use for specific scenarios. These
    #   are the standard error codes in Linux or UNIX.

    # some fake errors used for testing:
    TST_NEG_ERROR = -1
    TST_FLOAT_ERROR = 3.14159
    TST_STRING_NUMBER = "42"
    TST_SOME_ERROR = "SOME ERROR"

    ERR_ERROR_ = 1  # catchall for general errors
    ERR_SHELLERR_ = 2  # misuse of shell builtins (BASH)

    ERR__BASE_ = 64  # base value for error messages
    ERR_USAGE_ = 64  # command line usage error
    ERR_DATAERR_ = 65  # data format error
    ERR_NOINPUT_ = 66  # cannot open input
    ERR_NOUSER_ = 67  # addressee unknown
    ERR_NOHOST_ = 68  # host name unknown
    ERR_UNAVAILABL_ = 69  # service unavailable
    ERR_SOFTWARE_ = 70  # internal software error
    ERR_OSERR_ = 71  # system error (e.g., can't fork)
    ERR_OSFILE_ = 72  # critical OS file missing
    ERR_CANTCREAT_ = 73  # can't create (user) output file
    ERR_IOERR_ = 74  # input/output error
    ERR_TEMPFAIL_ = 75  # temp failure; user is invited to retry
    ERR_PROTOCOL_ = 76  # remote error in protocol
    ERR_NOPERM_ = 77  # permission denied
    ERR_CONFIG_ = 78  # configuration error
    ERR__MAX_ = 78  # maximum listed value

    # Linux / Unix codes
    ERR_CANTEXECUTE_ = 126  # command invoked cannot execute
    ERR_NOTFOUND_ = 127  # command not found; possible $PATH error or typo
    ERR_BADARG_ = 128  # invalid argument
    ERR_FATALARG_ = 129  # fatal error
    ERR_CTRL_C_ = 130  # script terminated by Control-C

    def __str__(self):
        a = self.value
        return str(f"Error: {self.name} (code: {self.value})")

    def err_msg(self):
        retval = f"Error "


def path_parts(p):
    return p.split("/")


def test_err():
    # some fake errors used for testing:
    # TST_NEG_ERROR = -1
    # TST_FLOAT_ERROR = 3.14159
    # TST_STRING_NUMBER = '42'
    # TST_SOME_ERROR = 'SOME ERROR'

    script_name = argv[0]
    name = path_parts(script_name)[-1]
    print(name)
    print(f"Testing {name} error codes ...\n\n")

    err_list = ["Err(72)", "Err."]
    errno = Err(72)

    print(errno)
    print(errno.ERR_IOERR_)

    for m in Enum.__dir__(errno):
        # print(m)
        detail = f"enum.Enum.{m}"
        s = f"{detail}"
        print(s)


def main():
    """
    CLI script main entry point.
    """
    print()
    print(FLAGS_ARGS)
    print()
    print(is_arg("--test", "test"))
    print()

    args = argv[1:]
    for arg in args:
        if is_arg(arg, "debug"):
            _DEBUG = True
            args.pop(arg)
        if is_arg(arg, "help"):
            _HELP = True
            args.pop(arg)
        if is_arg(arg, "test"):
            _RUN_TEST = True
            args.pop(arg)
        if is_arg(arg, "version"):
            _RUN_TEST = True
            args.pop(arg)
    if _RUN_TEST:
        test_err()
    else:
        for arg in args:
            s = f"Err({arg})"
            t = eval(s)
            u = f"{t:20}"
            print(s)
            print(u)
            return f"{s}"


if __name__ == "__main__":  # if script is loaded directly from CLI
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# 'Standard Library'
import os

from time import *

from typing import List


def fib_basic(n: int) -> (int, int):
    if n < 2:
        return n
    return fib_basic(n - 1) + fib_basic(n - 2)


if __name__ == "__main__":
    for i in range(20):
        print(i, fib_basic(i))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
""" text_colors version 0.9.3

    The obligatory ANSI text colors implementation.
    Requires Python 3.6+
    """

# 'Standard Library'
import json
import locale
import os
import sys

# 'package imports'
from autosys import __version__ as version
from autosys.bak.system import py_shell, pyver

from typing import Any, Dict, List

# locale.setlocale(locale.LC_ALL, '')
DEFAULT_ENCODING = locale.getpreferredencoding(do_setlocale=True)

# Initialize constants, variables, strings, etc...
ANSI_CHART = """ANSI Effects Chart
    <code>
    ╔══════════╦════════════════════════════════╦═════════════════════════════════════════════════════════════════════════╗
    ║  Code    ║             Effect             ║                                   Note                                  ║
    ╠══════════╬════════════════════════════════╬═════════════════════════════════════════════════════════════════════════╣
    ║ 0        ║  Reset / Normal                ║  all attributes off                                                     ║
    ║ 1        ║  Bold or increased intensity   ║                                                                         ║
    ║ 2        ║  Faint (decreased intensity)   ║  NWS                                                  ║
    ║ 3        ║  Italic                        ║  NWS Sometimes treated as inverse.                    ║
    ║ 4        ║  Underline                     ║                                                                         ║
    ║ 5        ║  Slow Blink                    ║  less than 150 per minute                                               ║
    ║ 6        ║  Rapid Blink                   ║  MS-DOS ANSI.SYS; 150+ per minute; not widely supported                 ║
    ║ 7        ║  [[reverse video]]             ║  swap foreground and background colors                                  ║
    ║ 8        ║  Conceal                       ║  NWS                                                  ║
    ║ 9        ║  Crossed-out                   ║  Characters legible, but marked for deletion.  NWS    ║
    ║ 10       ║  Primary(default) font         ║                                                                         ║
    ║ 11–19    ║  Alternate font                ║  Select alternate font `n-10`                                           ║
    ║ 20       ║  Fraktur                       ║  hardly ever supported                                                  ║
    ║ 21       ║  Bold off or Double Underline  ║  Bold off not widely supported; double underline hardly ever supported. ║
    ║ 22       ║  Normal color or intensity     ║  Neither bold nor faint                                                 ║
    ║ 23       ║  Not italic, not Fraktur       ║                                                                         ║
    ║ 24       ║  Underline off                 ║  Not singly or doubly underlined                                        ║
    ║ 25       ║  Blink off                     ║                                                                         ║
    ║ 27       ║  Inverse off                   ║                                                                         ║
    ║ 28       ║  Reveal                        ║  conceal off                                                            ║
    ║ 29       ║  Not crossed out               ║                                                                         ║
    ║ 30–37    ║  Set foreground color          ║  See color table below                                                  ║
    ║ 38       ║  Set foreground color          ║  Next arguments are `5;n` or `2;r;g;b`, see below                       ║
    ║ 39       ║  Default foreground color      ║  implementation defined (according to standard)                         ║
    ║ 40–47    ║  Set background color          ║  See color table below                                                  ║
    ║ 48       ║  Set background color          ║  Next arguments are `5;n` or `2;r;g;b`, see below                       ║
    ║ 49       ║  Default background color      ║  implementation defined (according to standard)                         ║
    ║ 51       ║  Framed                        ║                                                                         ║
    ║ 52       ║  Encircled                     ║                                                                         ║
    ║ 53       ║  Overlined                     ║                                                                         ║
    ║ 54       ║  Not framed or encircled       ║                                                                         ║
    ║ 55       ║  Not overlined                 ║                                                                         ║
    ║ 60       ║  ideogram underline            ║  hardly ever supported                                                  ║
    ║ 61       ║  ideogram double underline     ║  hardly ever supported                                                  ║
    ║ 62       ║  ideogram overline             ║  hardly ever supported                                                  ║
    ║ 63       ║  ideogram double overline      ║  hardly ever supported                                                  ║
    ║ 64       ║  ideogram stress marking       ║  hardly ever supported                                                  ║
    ║ 65       ║  ideogram attributes off       ║  reset the effects of all of 60-64                                      ║
    ║ 90–97    ║  Set bright foreground color   ║  aixterm (not in standard)                                              ║
    ║ 100–107  ║  Set bright background color   ║  aixterm (not in standard)                                              ║
    ╚══════════╩════════════════════════════════╩═════════════════════════════════════════════════════════════════════════╝
    </code>
    """

#  38 - Set foreground color - Next arguments are `5;n` or `2;r;g;b`, see below
ANSI_DICT_LIST = [
    {"code": "Code", "effect": "Effect", "notes": "Note"},
    {"code": "0", "effect": "Reset/Normal", "notes": "allattributesoff"},
    {"code": "1", "effect": "Boldorincreasedintensity", "notes": ""},
    {"code": "2", "effect": "Faint(decreasedintensity)", "notes": "NWS"},
    {
        "code": "3",
        "effect": "Italic",
        "notes": "NWSSometimestreatedasinverse.",
    },
    {"code": "4", "effect": "Underline", "notes": ""},
    {"code": "5", "effect": "SlowBlink", "notes": "lessthan150perminute"},
    {
        "code": "6",
        "effect": "RapidBlink",
        "notes": "MS-DOSANSI.SYS;150+perminute;notwidelysupported",
    },
    {
        "code": "7",
        "effect": "[[reversevideo]]",
        "notes": "swapforegroundandbackgroundcolors",
    },
    {"code": "8", "effect": "Conceal", "notes": "NWS"},
    {
        "code": "9",
        "effect": "Crossed-out",
        "notes": "Characterslegible,butmarkedfordeletion.NWS",
    },
    {"code": "10", "effect": "Primary(default)font", "notes": ""},
    {
        "code": "11–19",
        "effect": "Alternatefont",
        "notes": "Selectalternatefont`n-10`",
    },
    {"code": "20", "effect": "Fraktur", "notes": "hardlyeversupported"},
    {
        "code": "21",
        "effect": "BoldofforDoubleUnderline",
        "notes": "Boldoffnotwidelysupported;doubleunderlinehardlyeversupported.",
    },
    {
        "code": "22",
        "effect": "Normalcolororintensity",
        "notes": "Neitherboldnorfaint",
    },
    {"code": "23", "effect": "Notitalic,notFraktur", "notes": ""},
    {
        "code": "24",
        "effect": "Underlineoff",
        "notes": "Notsinglyordoublyunderlined",
    },
    {"code": "25", "effect": "Blinkoff", "notes": ""},
    {"code": "27", "effect": "Inverseoff", "notes": ""},
    {"code": "28", "effect": "Reveal", "notes": "concealoff"},
    {"code": "29", "effect": "Notcrossedout", "notes": ""},
    {
        "code": "30",
        "effect": "Setforegroundcolor",
        "notes": "BLACK, add ;1m GRAY",
    },
    {
        "code": "31",
        "effect": "Setforegroundcolor",
        "notes": "RED, add ;1m BRIGHTRED",
    },
    {
        "code": "32",
        "effect": "Setforegroundcolor",
        "notes": "GREEN, add ;1m BRIGHTGREEN",
    },
    {
        "code": "33",
        "effect": "Setforegroundcolor",
        "notes": "YELLOW, add ;1m BRIGHTYELLOW",
    },
    {
        "code": "34",
        "effect": "Setforegroundcolor",
        "notes": "BLUE, add ;1m BRIGHTBLUE",
    },
    {
        "code": "35",
        "effect": "Setforegroundcolor",
        "notes": "MAGENTA, add ;1m BRIGHTMAGENTA",
    },
    {
        "code": "36",
        "effect": "Setforegroundcolor",
        "notes": "CYAN, add ;1m BRIGHTCYAN",
    },
    {
        "code": "37",
        "effect": "Setforegroundcolor",
        "notes": "LIGHTGRAY, add ;1m WHITE",
    },
    # "BLACK": "\u001b[30m",
    # "RED": "\u001b[31m",
    # "GREEN": "\u001b[32m",
    # "BROWN": "\u001b[33m",
    # "BLUE": "\u001b[34m",
    # "MAGENTA": "\u001b[35m",
    # "CYAN": "\u001b[36m",
    # "GRAY": "\u001b[37m",
    # "BBLACK": "\u001b[30;1m",
    # "BRED": "\u001b[31;1m",
    # "BGREEN": "\u001b[32;1m",
    # "BYELLOW": "\u001b[33;1m",
    # "BBLUE": "\u001b[34;1m",
    # "BMAGENTA": "\u001b[35;1m",
    # "BCYAN": "\u001b[36;1m",
    # "BWHITE": "\u001b[37;1m",
    {
        "code": "38",
        "effect": "Setforegroundcolor",
        "notes": "Nextargumentsare`5;n`or`2;r;g;b`,seebelow",
    },
    {
        "code": "39",
        "effect": "Defaultforegroundcolor",
        "notes": "implementationdefined(accordingtostandard)",
    },
    {
        "code": "40–47",
        "effect": "Setbackgroundcolor",
        "notes": "Seecolortablebelow",
    },
    {
        "code": "48",
        "effect": "Setbackgroundcolor",
        "notes": "Nextargumentsare`5;n`or`2;r;g;b`,seebelow",
    },
    {
        "code": "49",
        "effect": "Defaultbackgroundcolor",
        "notes": "implementationdefined(accordingtostandard)",
    },
    {"code": "51", "effect": "Framed", "notes": ""},
    {"code": "52", "effect": "Encircled", "notes": ""},
    {"code": "53", "effect": "Overlined", "notes": ""},
    {"code": "54", "effect": "Notframedorencircled", "notes": ""},
    {"code": "55", "effect": "Notoverlined", "notes": ""},
    {
        "code": "60",
        "effect": "ideogramunderline",
        "notes": "hardlyeversupported",
    },
    {
        "code": "61",
        "effect": "ideogramdoubleunderline",
        "notes": "hardlyeversupported",
    },
    {
        "code": "62",
        "effect": "ideogramoverline",
        "notes": "hardlyeversupported",
    },
    {
        "code": "63",
        "effect": "ideogramdoubleoverline",
        "notes": "hardlyeversupported",
    },
    {
        "code": "64",
        "effect": "ideogramstressmarking",
        "notes": "hardlyeversupported",
    },
    {
        "code": "65",
        "effect": "ideogramattributesoff",
        "notes": "resettheeffectsofallof60-64",
    },
    {
        "code": "90–97",
        "effect": "Setbrightforegroundcolor",
        "notes": "aixterm(notinstandard)",
    },
    {
        "code": "100–107",
        "effect": "Setbrightbackgroundcolor",
        "notes": "aixterm(notinstandard)",
    },
]


# def print_ansi_list():
#     for line in ANSI_DICT_LIST:


FG_DICT = {
    # The first 11 codes are just favorites I picked.
    # Some have short and long names for duplicates.
    # Add more personalized colors by name if desired ...
    "MAIN": "\u001b[38;5;229m",
    "COOL": "\u001b[38;5;38m",
    "MIKES_BLUE": "\u001b[38;5;38m",
    "GO": "\u001b[38;5;28m",
    "GREEN_MEANS_GO": "\u001b[38;5;28m",
    "WARN": "\u001b[38;5;203m",
    "CHERRY": "\u001b[38;5;124m",
    "CANARY": "\u001b[38;5;226m",
    "DANGERWILLROBINSON": "\u001b[38;5;203m",
    "ATTN": "\u001b[38;5;178m",
    "WHATABURGER": "\u001b[38;5;178m",
    "PURPLE": "\u001b[38;5;93m",
    "PURPLEHAZE": "\u001b[38;5;93m",
    "RESET": "\u001b[0m",  # Reset all attributes
    "NEXT_FG": "\u001b[38m",  # Next Foreground in Cycle (rainbow ...)
    "DEF_FG": "\u001b[39m",  # Reset to Default Foreground
    # These are the codes for the foreground 8 color set
    "BLACK": "\u001b[30m",  # 0
    "RED": "\u001b[31m",  # 1
    "GREEN": "\u001b[32m",  # 2
    "BROWN": "\u001b[33m",  # 3
    "BLUE": "\u001b[34m",  # 4
    "MAGENTA": "\u001b[35m",  # 5
    "CYAN": "\u001b[36m",  # 6
    "GRAY": "\u001b[37m",  # 7
    # These are added to the 8-color FG to make the 16 color FG set
    "BBLACK": "\u001b[30;1m",
    "BRED": "\u001b[31;1m",
    "BGREEN": "\u001b[32;1m",
    "BYELLOW": "\u001b[33;1m",
    "BBLUE": "\u001b[34;1m",
    "BMAGENTA": "\u001b[35;1m",
    "BCYAN": "\u001b[36;1m",
    "BWHITE": "\u001b[37;1m",
}

BG_DICT = {
    # The first 11 codes are just favorites I picked.
    # Some have short and long names for duplicates.
    # Add more personalized colors by name if desired ...
    "MAIN": "\u001b[48;5;185m",
    "COOL": "\u001b[48;5;38m",
    "MIKES_BLUE": "\u001b[48;5;38m",
    "GO": "\u001b[48;5;28m",
    "GREEN_MEANS_GO": "\u001b[48;5;28m",
    "WARN": "\u001b[48;5;203m",
    "CHERRY": "\u001b[48;5;124m",
    "CANARY": "\u001b[48;5;226m",
    "DANGERWILLROBINSON": "\u001b[48;5;203m",
    "ATTN": "\u001b[48;5;178m",
    "WHATABURGER": "\u001b[48;5;178m",
    "PURPLE": "\u001b[48;5;93m",
    "PURPLEHAZE": "\u001b[48;5;93m",
    "RESET": "\u001b[0m",  # Reset all attributes
    "NEXT_BG": "\u001b[48m",  # Next Background in Cycle (rainbow ...)
    "DEF_BG": "\u001b[49m",  # Reset to Default Background
    # These are the codes for the background 8 color set
    "BG_BLACK": "\u001b[40m",
    "BG_RED": "\u001b[41m",
    "BG_GREEN": "\u001b[42m",
    "BG_BROWN": "\u001b[43m",
    "BG_BLUE": "\u001b[44m",
    "BG_MAGENTA": "\u001b[45m",
    "BG_CYAN": "\u001b[46m",
    "BG_GRAY": "\u001b[47m",
    # These are added to the 8-color BG to make the 16 color BG set
    "BG_BBlack": "\u001b[40;1m",
    "BG_BRed": "\u001b[41;1m",
    "BG_BGreen": "\u001b[42;1m",
    "BG_BYellow": "\u001b[43;1m",
    "BG_BBlue": "\u001b[44;1m",
    "BG_BMagenta": "\u001b[45;1m",
    "BG_BCyan": "\u001b[46;1m",
    "BG_BWhite": "\u001b[47;1m",
}

FLAGS_DICT = {
    "RESET": "\u001b[0m",
    "BOLD": "\u001b[1m",
    "DIM": "\u001b[2m",
    "ITALIC": "\u001b[3m",
    "UNDER": "\u001b[4m",
    "REVERSE": "\u001b[7m",
    "CONCEAL": "\u001b[8m",
}


def _generate_ansi256_codes():
    """
    Fill in the 256 set color codes
    """
    #   => (0 to 231) colors in the format 'COLOR133'
    tmp_color_code: int = 0
    for tmp_color_code in range(0, 232):
        FG_DICT["COLOR" + str(tmp_color_code)] = (
            "\u001b[38;5;" + str(tmp_color_code) + "m"
        )
        BG_DICT["COLOR" + str(tmp_color_code)] = (
            "\u001b[48;5;" + str(tmp_color_code) + "m"
        )

    #   => (232 to 255) grayscale in the format 'GRAY243'
    for tmp_color_code in range(232, 256):
        key = "GRAY" + str(tmp_color_code)
        FG_DICT["GRAY" + str(tmp_color_code)] = (
            "\u001b[38;5;" + str(tmp_color_code) + "m"
        )
        BG_DICT["GRAY" + str(tmp_color_code)] = (
            "\u001b[48;5;" + str(tmp_color_code) + "m"
        )


_generate_ansi256_codes()

# Turn off color when using ipython, otherwise leave it on
ENABLE_COLOR: bool = True
if py_shell() in ["ipython", "ipython-notebook"]:
    ENABLE_COLOR = False

print(f"{py_shell()=}")
print(f"{ENABLE_COLOR=}")
DEFAULT_FG_COLOR: str = "MAIN" * ENABLE_COLOR
DEFAULT_BG_COLOR: str = "BG_BLACK" * ENABLE_COLOR
DEFAULT_FLAGS: str = "RESET" * ENABLE_COLOR


def color_encode(
    fg_color: str = DEFAULT_FG_COLOR,
    bg_color: str = DEFAULT_BG_COLOR,
    flags_color: str = DEFAULT_FLAGS,
) -> str:
    """ Returns encoded ANSI text string from components.

        Keyword Arguments:
                fg_color {str} - ANSI foreground color {DEFAULT_FG_COLOR}
                bg_color {str} - ANSI background color {DEFAULT_BG_COLOR}
                flags_color {str} - ANSI format codes

        Returns:
            str - printable escaped ANSI text formatting string
        """
    color_code: str = ""
    if ENABLE_COLOR:
        color_code += FG_DICT[fg_color]
        color_code += BG_DICT[bg_color]
        color_code += FLAGS_DICT[flags_color]
    return color_code


DEFAULT_ENCODE: str = color_encode(
    DEFAULT_FG_COLOR, DEFAULT_BG_COLOR, DEFAULT_FLAGS
)
STICKY_ENCODE: str = DEFAULT_ENCODE
# This code will reset all ANSI codes to default.
RESET_ENCODE: str = color_encode(
    DEFAULT_FG_COLOR, DEFAULT_BG_COLOR, DEFAULT_FLAGS
)

USAGE: str = f"""

author    - Michael Treanor  <skeptycal@gmail.com>
copyright - 2019 (c) Michael Treanor
license   - MIT <https://opensource.org/licenses/MIT>
github    - https://www.github.com/skeptycal

Usage: {os.path.basename(__file__)} [demo|version|help]

  Parameters:
      demo, -d, --demo        -- install and initialize
      version, -v, --version  -- display version information
      help, -h, --help        -- display usage and information
    """
# *############## Print Functions


def color_print(color_code: str, *args: Any, **kwargs: Any) -> int:
    """ Print using ANSI color codes

        Arguments:
            color_code {str} - printable escaped ANSI text formatting string
            args {tuple} - passthrough for print arguments
            kwargs {dict} - passthrough for print options

        Return:
            int: 0 for success; > 0 for error code """
    try:
        if ENABLE_COLOR:
            print(color_code, *args, STICKY_ENCODE, **kwargs)
        else:
            print(*args, **kwargs)
    except RuntimeError:
        return 1
    return 0


def color_gradient(color_1: int, color_2: int, print_string: str) -> int:
    """ Print <print_string> using a calculated ANSI color gradient
            that changes evenly and gradually between color_1 and color_2.

        Arguments:
            color_1 {int} - ANSI color code(256 set) - lower gradient bound
            color_2 {int} - ANSI color code(256 set) - upper gradient bound
            print_string {str} - target string

        Returns:
            int: 0 for success; > 0 for error code """
    if ENABLE_COLOR and len(print_string) >= 4:
        result: str = ""
        # make sure color_2 is the larger number
        if color_1 > color_2:
            color_1, color_2 = color_2, color_1
        try:
            gradient_slope: float = (color_2 - color_1) / len(print_string)
            for i, _ in enumerate(print_string):
                color_code = "COLOR" + str(int(color_1 + i * gradient_slope))
                result += color_encode(color_code) + print_string[i]
            print(result)
        except TypeError:
            return 1
    return 0


def color_cycle(color_1: int, print_string: str) -> int:
    """ Print by cycling through colors starting at <color_1>.

        Arguments:
            color_1 {int} - ANSI color code(256 set) - lower gradient bound
            print_string {str} - target string

        Returns:
            int: 0 for success; > 0 for error code """
    if ENABLE_COLOR:
        try:
            for _, c in enumerate(print_string):
                color_1 += 1
                if color_1 < 1 or color_1 > 230:
                    color_1 = 1
                color_code = "COLOR" + str(color_1)
                print(color_encode(color_code), str(c), sep="", end="")
            print(DEFAULT_ENCODE, " ", sep="", end="")
        except (RuntimeError, IndexError):
            return 1
    return 0


def loading(count: int) -> int:
    """ Demo - Print <count> text progress bars.
            They are non responsive and synchronous.

        Arguments:
            count {int} - number of bars

        Returns:
            int: 0 for success; > 0 for error code """
    import time
    import random

    width: int = 0
    w_string: str = ""
    all_progress = [0] * count
    sys.stdout.write("\n" * count)  # space to draw the bars
    while any(x < 100 for x in all_progress):
        time.sleep(0.003)
        # Randomly increment one of our progress values
        unfinished = [(i, v) for (i, v) in enumerate(all_progress) if v < 100]
        index, _ = random.choice(unfinished)
        all_progress[index] += 1
        # Draw the progress bars
        sys.stdout.write("\u001b[1000D")  # Move left
        sys.stdout.write("\u001b[" + str(count) + "A")  # Move up
        for progress in all_progress:
            width = int(progress / 4)
            w_string = "#" * width
            pad_string = " " * (25 - width)
            if len(w_string) >= 5:
                pad_string = " " * (25 - width)
                pad_string = "|" + w_string + pad_string + "]"
            else:
                pad_string = " " * 21
                pad_string = "|" + w_string + pad_string + "]"
            color_cycle(124, pad_string)
    return 0


def fg_samples():
    """ Foreground color samples """
    i: int = 0
    for k in FG_DICT:
        if i < 32:
            if not i % 8:
                print()
        elif not i % 6:
            print()
        i += 1
        color_print(
            color_encode(k, DEFAULT_BG_COLOR),
            k.replace("OLOR", ""),
            " ",
            sep="",
            end="",
        )
    print()


def bg_samples():
    """ Background color samples """
    i = 0
    for k in BG_DICT:
        if i < 32:
            if not i % 8:
                print()
        elif not i % 6:
            print()
        i += 1
        color_print(
            color_encode(DEFAULT_FG_COLOR, k),
            k.replace("OLOR", ""),
            " ",
            sep="",
            end="",
        )
    print()


def flags_samples():
    """ ANSI Flags samples """
    print()
    i: int = 0
    s: str = "Now is the time for all good men" + "to come to the aid of their country."
    for k in FLAGS_DICT:
        i += 1
        color_print(
            color_encode("BWHITE", "PURPLEHAZE", k),
            i,
            " : ",
            str(k),
            "  \t",
            s,
            RESET_ENCODE,
            sep="",
        )
    print()
    color_gradient(130, 144, "Gradient!!!  \t{}\n".format(s))
    print()
    color_cycle(130, "Cycle!!!  \t{}\n".format(s * 5))
    print()
    # if ENABLE_COLOR:
    #     loading(5)
    # print()
    color_print(DEFAULT_ENCODE)
    color_print(
        color_encode("BWHITE", "PURPLEHAZE", "ITALIC"), "TESTS COMPLETE ..."
    )


if __name__ == "__main__":
    # TEST SAMPLES to use if script is run from the command lines

    if len(sys.argv) < 2:
        # color_cycle(42, usage)
        color_gradient(42, 222, USAGE)
        pass
    else:
        arg1: str = sys.argv[1].lower()
        if arg1 in ["version", "-v", "--version"]:
            color_cycle(88, version)
        elif arg1 in ["help", "-h", "--help"]:
            color_print(
                color_encode("BWHITE", "PURPLEHAZE", "ITALIC"), version
            )
            print()
            color_cycle(88, ANSI_CHART)
        elif arg1 in ["demo", "-d", "--demo"]:
            fg_samples()
            bg_samples()
            flags_samples()
        else:
            color_gradient(42, 222, USAGE)
            # color_cycle(42, usage)

    print("\n\n")
    print("python version (pyver): ", pyver())
    print("python shell (py_shell): ", py_shell())


# Resources:

# https://www.geeksforgeeks.org/print-colors-python-terminal/

# https://stackoverflow.com/questions/287871/print-in-terminal-with-colors
#   1,000,002 ways to do this ... but I still did it my own way ...

# https://en.wikipedia.org/wiki/ANSI_escape_code

# Sequences have different lengths.
#   All sequences start with ESC (27 / hex 0x1B / octal 033), followed
#   by a second byte in the range 0x40 - 0x5F (ASCII @A-Z[\\]^_)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
""" as_flask.py - basic Flask app
    Reference: https://stackoverflow.com/a/26853961
    ---
    Copyright (c) 2018 [Michael Treanor][2]

    AutoSys is licensed under the [MIT License][3]

    [1]: https://www.github.com/skeptycal/autosys
    [2]: https://www.twitter.com/skeptycal
    [3]: https://opensource.org/licenses/MIT
    """

if True:
    import flask
    from flask import current_app, request
    from random import random, choice, choices
    # from autosys.profiler.quicktime import QuickTime as qt

try:
    DEFAULT_ENCODING
except:
    from locale import getpreferredencoding
    DEFAULT_ENCODING: str = getpreferredencoding(do_setlocale=True) or "utf-8"

try:
    from ujson import dumps
    JSON_PARSER = "ujson"
except ImportError as e:
    from json import dumps
    JSON_PARSER = "json"


class FlaskJSONParameterError:
    "ultrajsonify behavior is undefined when passed both args and kwargs"


def ultrajsonify(*args, **kwargs):
    """This function reimplements ``flask.json.jsonify``
    using ``ujson.dumps`` instead of ``json.dumps``.
    """
    indent = 0
    ensure_ascii = current_app.config.get("JSON_AS_ASCII", True)
    mimetype = current_app.config.get("JSONIFY_MIMETYPE", "application/json")

    if (current_app.config["JSONIFY_PRETTYPRINT_REGULAR"]
            and not request.is_xhr):
        indent = 2

        # if (args and kwargs) or not(args and kwargs):
    if (args and kwargs):
        raise FlaskJSONParameterError

    data = args[0] if len(args) == 1 else args or kwargs

    return current_app.response_class(
        dumps(data, indent=indent, ensure_ascii=ensure_ascii),
        mimetype=mimetype,
    )


flask.json.jsonify = ultrajsonify


def random_binary(n=100, set_size=2):
    return [choices([0, 1], k=set_size) for _ in range(n)]


def random_binary_gen(n=10, set_size=2):
    yield (choices((0, 1), set_size) for _ in range(n))


def xor1(a, b):
    return bool(a) != bool(b)


def xor2(a, b):
    return bool(a) ^ bool(b)


def xor3(a, b):
    return (a + b) == 1


def xor4(*args):
    return any(args)


def xor5(*args, **kwargs):
    return (args and kwargs) or not(args and kwargs)


fake = []
fake = random_binary(n=100)

# print(fake)

for c in fake:
    a, b = c
    print(a, b)
    print(qt(xor1(a, b)))
    # timeit('xor1(a, b)')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
""" # TODO -- @update `AutoSys` package
        copyright (c) 2018 Michael Treanor
        https://www.github.com/skeptycal/autosys
        https://www.twitter.com/skeptycal

    `AutoSys` is licensed under the `MIT License
        `<https://opensource.org/licenses/MIT>`


    String CONSTANTS

        ord(NL):  10

        [ord(x) for x in STR_WHITESPACE]=[32, 9, 10, 13, 11, 12]

        STR_HEX='0123456789abcdefABCDEF'`

        STR_ALPHA='abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ'

        STR_ALPHANUMERIC='abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789'

        STR_PUNCTUATION='!"#$%&\'()*+,-./:;<=>?@[\\]^_`{|}~'

        STR_PRINTABLE='0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ!"#$%&\'()*+,-./:;<=>?@[\\]^_`{|}~ \t\n\r\x0b\x0c'

        [ord(x) for x in STR_PRINTABLE]=[48, 49, 50, 51, 52, 53, 54, 55, 56, 57, 97, 98, 99, 100, 101, 102, 103, 104, 105, 106, 107, 108, 109, 110, 111, 112, 113, 114, 115, 116, 117, 118, 119, 120, 121, 122, 65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77, 78, 79, 80, 81, 82, 83, 84, 85, 86, 87, 88, 89, 90, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 58, 59, 60, 61, 62, 63, 64, 91, 92, 93, 94, 95, 96, 123, 124, 125, 126, 32, 9, 10, 13, 11, 12]
        """

# * --------------------------------- Imports and Constants

# 'Standard Library'
import re

from dataclasses import Field, dataclass, field

# 'package imports'
from autosys import debug
from random_string import *
from strang import Strang, random_string

from typing import Final, List, Tuple

CASE_LIST: Tuple = ("upper", "lower", "title", "snake", "camel", "pascal")

DEFAULT_RE_FLAGS: Final[int] = re.MULTILINE | re.IGNORECASE

RE_VERSION: re.Pattern = re.compile(
    pattern=r'^__version__\s?=\s?[\'"]([^\'"]*)[\'"]',
    flags=DEFAULT_RE_FLAGS,
)

RE_REPLACE: Final[re.Pattern] = re.compile(pattern=r"[-\s]",
                                           flags=DEFAULT_RE_FLAGS)


class ReUtilsError(Exception):
    "An error occurred while processing regex strings with ReUtils."


__all__ = [
    "CASE_LIST",
    "DEFAULT_RE_FLAGS",
    "NL",
    "NUL",
    "RE_REPLACE",
    "RE_VERSION",
    "ReUtils",
    "ReUtilsError",
    "STR_ALPHA",
    "STR_ALPHANUMERIC",
    "STR_HEX",
    "STR_NAMES",
    "STR_PRINTABLE",
    "STR_PUNCTUATION",
    "STR_WHITESPACE",
    "linesep",
    "random_string",
    "re",
    "re_pip_safe_name",
    "split_it",
    "string",
    "sub_it",
]
if True:  # * --------------------------------- ReUtils Class

    @dataclass
    class ReUtils(Strang):
        """ Wrapper for common Python3 Regex Utilities. """

        string: str = r""

        def __post_init__(self):
            # strip whitespace from ends
            self.string.strip()
            # self.string = self.string.strip()
            # remove duplicate whitespace
            self.dedupe_whitespace()
            self._set_cases()

        def __str__(self):
            return self.string

        def _set_cases(self):
            """ Choose a case by mapping the prefix to the methods.

                e.g.
                ```
                CASE_LIST = ('upper', 'lower', 'title', 'snake', 'camel', 'pascal')

                _cases = {x: f"self.to_{x}_case" for x in CASE_LIST}
                ```
                """

            self._cases = {x: f"self.to_{x}_case" for x in CASE_LIST}

        def re_pip_safe_name(self) -> (str, None, Exception):
            """ Return a name that is converted to pypi safe format.

                Replace ' ' (space) and '-'(hyphen) with _(underscore) using python3 regex
                """
            try:
                return RE.sub_it(string=self.string.lower(),
                                 pattern=RE_REPLACE,
                                 repl="_")
            except Exception as e:
                raise ReUtilsError(e)

        def sub_it(self,
                   pattern: re.Pattern = RE_REPLACE,
                   repl: str = "_") -> (str):
            """ Return a string where items in `pattern` have been replaced with `repl`. """
            try:
                return re.sub(pattern=pattern, repl=repl, string=self.string)
            except Exception as e:
                raise ReUtilsError(e)

        def split_it(self, delimiter: str = "\s") -> (str):
            """ Return a string separated using 'delimiter' using python3 regex. """
            try:
                return re.split(pattern=delimiter, string=self.string)
            except Exception as e:
                raise ReUtilsError(e)

        def clear_all_whitespace(self):
            return self.string.translate(
                {ord(c): None
                 for c in string.whitespace})
            # TODO - which is faster?
            # self.string = "".join(self.string.split())

        def dedupe_whitespace(self):
            """ Return a string with duplicate whitespace replaced with ' '. """
            return re.sub(pattern=r"[\s+]", repl=" ", string=self.string)


def _test():
    test_list1: List[str] = [
        "this is a-test",
        "123_456 789-111",
        "(361) 773-2832",
        ";alskjdfpo82jn sdf83nf ",
    ]

    test_whitespace: List[str] = [
        "  spaces all      over   the    \n places     \t",
        "white\t\n\r\x0b\x0cspace",
        "   fasdf   ",
        "a   b  c     d  e f g hij k  ",
    ]

    test_list: List[str] = []
    test_list.extend(test_whitespace)

    for i in range(20):
        test_list.append(random_string(25))

    re_delimiter = r"[\s-]"
    delimiter = r" -"
    print()
    for item in test_list:
        print("---------------------------------")
        print(item)
        print("---------------------------------")
        s = ReUtils(item)
        print("s:        ", s)
        print("repr:     ", repr(s))
        print("split:    ", s.split_it())
        print("sub:      ", s.sub_it())
        print("upper:    ", s.to_upper_case)
        print("lower:    ", s.to_lower_case)
        print("title:    ", s.to_title_case)
        print("snake:    ", s.to_snake_case)
        print("kebab:    ", s.to_kebab_case)
        print("camel:    ", s.to_camel_case)
        print("pascal:   ", s.to_pascal_case)
        print("clear:    ", s.clear_all_whitespace())


if True == False:  # * --------------------------------- Tests
    _test()



import requests
import lxml
from bs4 import BeautifulSoup as soup

data = requests.get(
    'https://www.signalboosters.com/blog/best-smartphone-apps-to-find-your-cell-signal-strength/')


soup.
soup.get_text()



#!/usr/bin/env python3
# -*- encoding: utf-8 -*-
""" ### Various temporary file implementations:

    - TMP: 'secure' temp file with no directory entry
    - TMP_NAMED: temp file with name that can be used in code
    - TMP_HERE: temp file with name; created in the current directory
    """

# 'Standard Library'
import enum
import os
import sys
import tempfile

# 'package imports'
import autosys.parsing.file_ops.file_class

# from io import StringIO
# from pathlib import Path



# !------------------------------------- Dev and Debugging
_debug_: bool = True  # Turn on extra debug info display

_NAME: str = os.path.basename(__file__)
_HERE: str = os.path.dirname(__file__)
_TMP_PREFIX: str = f"_{_NAME}_".replace(".", "_").strip().lower()
_TMP_SUFFIX: str = ".log"


class TempFile:
    TMP = tempfile.TemporaryFile()  # e.g. with TMP as f:
    TMP_NAMED = tempfile.NamedTemporaryFile()  # e.g. with TMP_NAMED as f:
    TMP_HERE = tempfile.NamedTemporaryFile(
        prefix=_TMP_PREFIX, suffix=_TMP_SUFFIX, dir=_HERE
    )
    TMP_SPOOL = tempfile.SpooledTemporaryFile(
        max_size=1024, prefix=_TMP_PREFIX, suffix=_TMP_SUFFIX, dir=_HERE
    )
    TMP_DIR = tempfile.TemporaryDirectory(
        prefix=_TMP_PREFIX, suffix=_TMP_SUFFIX
    )

    def __init__(self, temp_type: str = "TMP"):
        super().__init__()


def tmp_file(temp_type: str = "TMP"):
    try:
        return TempFile(f"{temp_type}")
    except Exception as e:
        return e


def temp_it(func):
    """ Wrapper for temp files. """
    pass


__all__ = ["TMP", "TMP_NAMED", "TMP_HERE"]

# !------------------------------------- Dev and Debugging


def _test_():
    """ #### Perform tests to produce additional debug info.

        (runs if _debug_ = True)
        """

    print(f"{_NAME=}")
    print(f"{_HERE=}")
    print(f"{_TMP_PREFIX=}")
    print(f"{_TMP_SUFFIX=}")
    print()

    script_name: str = os.path.basename(__file__)
    print(f"{script_name} is in DEBUG MODE.")
    print(TempFile.TMP)
    print(TempFile.TMP_HERE)
    print(TempFile.TMP_NAMED)
    print(TempFile.TMP_SPOOL)
    print(tmp_file())
    with TempFile.TMP as f:
        print_file_info(f)
    # print('exists after: ', os.path.exists(f.name))

    # with TMP_NAMED as f:
    #     print_file_info(f)
    # print('exists after: ', os.path.exists(f.name))

    # with TMP_HERE as f:
    #     print_file_info(f)
    # print('exists after: ', os.path.exists(f.name))

    print()

    print("\n --------------- done ...")


# !------------------------------------- main entry point


def _main_():
    """
    CLI script main entry point.
    """
    if _debug_:
        _test_()  # CLI testing - only runs if _DEBUG_ = True


if __name__ == "__main__":  # if script is loaded directly from CLI
    _main_()  # CLI functionality



#!/usr/bin/env python

# Copyright 2007-2016 The Python-Twitter Developers

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# ----------------------------------------------------------------------
# Change History
#
# 2010-05-16
#   TinyURL example and the idea for this comes from a bug filed by
#   acolorado with patch provided by ghills.  Class implementation
#   was done by @bear.
#
#   Issue #19: http://code.google.com/p/python-twitter/issues/detail?id=19
#
# 2016-02-18
#   Updated example with code to demonstrate passing a status message through
#   a shortener and then off to PostUpdate. Implemenation by @jeremylow from
#   bug filed by @immanuelfactor
#
#   Issue #298: https://github.com/bear/python-twitter/issues/298

# ----------------------------------------------------------------------
# This file demonstrates how to shorten all URLs contained within a Tweet
# by passing the tweet text to a shortener. In this case, we're using TinyURL
# since it does not require any real authentication for our purposes. If you
# are using a different service to shorten URLs, then you will need to modify
# the ShortenURL class to suit your needs.

# Note that this example shortens all URLs contained within the Tweet text.

# To use this example, replace the W/X/Y/Zs with your keys obtained from
# Twitter, or uncomment the lines for getting an environment variable. If you
# are using a virtualenv on Linux, you can set environment variables in the
# ~/VIRTUALENVDIR/bin/activate script.

# If you need assistance with obtaining keys from Twitter, see the instructions
# in doc/getting_started.rst.


# 'Standard Library'
import re

# 'package imports'
from twitter import Api
from twitter.twitter_utils import URL_REGEXP

try:
    from urllib.request import urlopen
except:
    from urllib2 import urlopen



class ShortenURL(object):
    """ A class that defines the default URL Shortener.

    TinyURL is provided as the default and as an example helper class to make
    URL Shortener calls if/when required. """

    def __init__(self, userid=None, password=None):
        """Instantiate a new ShortenURL object. TinyURL, which is used for this
        example, does not require a userid or password, so you can try this
        out without specifying either.

        Args:
            userid:   userid for any required authorization call [optional]
            password: password for any required authorization call [optional]
        """
        self.userid = userid
        self.password = password

    def Shorten(self, long_url):
        """ Call TinyURL API and returned shortened URL result.

        Args:
            long_url: URL string to shorten

        Returns:
            The shortened URL as a string

        Note:
            long_url is required and no checks are made to ensure completeness
        """

        result = None
        f = urlopen(
            "http://tinyurl.com/api-create.php?url={0}".format(long_url)
        )
        try:
            result = f.read()
        finally:
            f.close()

        # The following check is required for py2/py3 compatibility, since
        # urlopen on py3 returns a bytes-object, and urlopen on py2 returns a
        # string.
        if isinstance(result, bytes):
            return result.decode("utf8")
        else:
            return result


def _get_api():
    # Either specify a set of keys here or use os.getenv('CONSUMER_KEY') style
    # assignment:

    CONSUMER_KEY = "WWWWWWWW"
    # CONSUMER_KEY = os.getenv("CONSUMER_KEY", None)
    CONSUMER_SECRET = "XXXXXXXX"
    # CONSUMER_SECRET = os.getenv("CONSUMER_SECRET", None)
    ACCESS_TOKEN = "YYYYYYYY"
    # ACCESS_TOKEN = os.getenv("ACCESS_TOKEN", None)
    ACCESS_TOKEN_SECRET = "ZZZZZZZZ"
    # ACCESS_TOKEN_SECRET = os.getenv("ACCESS_TOKEN_SECRET", None)

    return Api(
        CONSUMER_KEY, CONSUMER_SECRET, ACCESS_TOKEN, ACCESS_TOKEN_SECRET
    )


def PostStatusWithShortenedURL(status):
    shortener = ShortenURL()
    api = _get_api()

    # Find all URLs contained within the status message. Value of ``urls`` will
    # be a list.
    urls = re.findall(URL_REGEXP, status)

    for url in urls:
        status = status.replace(url, shortener.Shorten(url), 1)

    api.PostUpdate(status)


if __name__ == "__main__":
    PostStatusWithShortenedURL("this is a test: http://www.example.com/tests")



# Profiling examples


# List.reverse()
# vs
# List[::-1]

# dict.get
# vs
# dict[""]

# finding factors:
# for i in range(1, f + 1):
#   if f % i == 0:


# finding memory usage:
# sys.getsizeof()
# vs
# memoryview
# vs
# memory stream


# clear screen
# def screen_clear():
#    # for mac and linux(here, os.name is 'posix')
#    if os.name == 'posix':
#       _ = os.system('clear')
#    else:
#       # for windows platfrom
#       _ = os.system('cls')
#    # print out some text
# vs
# terminal CSI commands



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
""" Part of the `AutoSys` package
        copyright (c) 2019 Michael Treanor
        https://www.github.com/skeptycal/autosys
        https://www.twitter.com/skeptycal

    `AutoSys` is licensed under the `MIT License
        `<https://opensource.org/licenses/MIT>`
    """

# Nice example from stack overflow
# reference: https://codereview.stackexchange.com/a/245715

# Core Library modules
import operator
import random
import timeit
from dataclasses import dataclass

# Third party modules
import matplotlib.pyplot as plt  # NOQA
import numpy as np
import seaborn as sns  # NOQA


def is_palindrome_str(x: int) -> bool:
    s = str(x)

    if len(s) == 1:
        palindrome = True

    i = int(len(s) / 2)

    for p in range(i):
        if s[p] != s[-(p + 1)]:
            return False
    # this is wrong ... it sets palindrome to True if ANY char matches
    # else:
    #     palindrome = True
    return True


def is_palindrome_str_refactored(x: int) -> bool:
    s = str(x)

    length = len(s)
    if length == 1:
        return True

    # i = int(len(string) / 2) # not sure what this was for ...
    i = length // 2

    for p in range(i):
        if s[p] != s[-(p + 1)]:
            return False

    return True


def is_palindrome_dict(x: int) -> bool:
    s = {i: j for i, j in enumerate(str(x))}
    i = (max(s) + 1) // 2

    if max(s) + 1 == 1:
        return True

    for p in range(i):
        if s[p] != s[max(s) - p]:
            return False
        else:
            palindrome = True

    return palindrome


def quick_check(x: int) -> bool:
    """ A very efficient check to benchmark against. """
    string = str(int)

    length = len(string)
    if length == 1:
        return True

    mid = length // 2

    start = string[:mid]
    finish = string[-mid:]

    # print(f"{start}  ...  {finish}")
    if start == finish:
        return True
    return False


def generate_palindrome(nb_chars):
    chars = "abcdefghijklmnopqrstuvwxyz"
    prefix = "".join([random.choice(chars) for _ in range(nb_chars)])
    mid = random.choice(["", "x"])
    return prefix + mid + prefix[::-1]


def generate_not_palindrome(nb_chars):
    chars = "abcdefghijklmnopqrstuvwxyz"
    prefix = "".join([random.choice(chars) for _ in range(nb_chars)])
    mid = random.choice(["", "x"])
    # these two are intentionally set different and in the middle of the string
    # TODO - could place errors in random locations?
    fudge = "x"
    caramel = "y"
    return prefix + caramel + mid + fudge + prefix[::-1]


def create_boxplot(duration_list):
    plt.figure(num=None, figsize=(8, 4), dpi=300, facecolor="w", edgecolor="k")
    sns.set(style="whitegrid")
    sorted_keys, sorted_vals = zip(
        *sorted(duration_list.items(), key=operator.itemgetter(1))
    )
    flierprops = dict(markerfacecolor="0.75", markersize=1, linestyle="none")
    ax = sns.boxplot(data=sorted_vals, width=0.3, orient="h", flierprops=flierprops,)
    ax.set(xlabel="Time in s", ylabel="")
    plt.yticks(plt.yticks()[0], sorted_keys)
    plt.tight_layout()
    plt.savefig("output.png")


def time_text_functions(functions, text, repeat=100, n=100, default_return=True):
    # palindromes...
    for func, name in functions:

        assert func(text) == default_return

        durations = timeit.repeat(lambda: func(text), repeat=repeat, number=n)
        duration_list[name] = durations
        print(
            "{func:<20}: "
            "min: {min:0.3f}s, mean: {mean:0.3f}s, max: {max:0.3f}s".format(
                func=name,
                min=min(durations),
                mean=np.mean(durations),
                max=max(durations),
            )
        )


@dataclass
class ProfileFunc:
    """ Profile a list of functions that accept text input.
    """

    DEFAULT_TEXT_LEN: int = 10
    n: int = 100
    repeat: int = 100

    def __init__(self):

        functions = functions[::-1]
        duration_list = {}


def main():

    text = generate_palindrome(nb_chars=string_len)
    not_text = generate_not_palindrome(nb_chars=string_len)

    # print(text)
    # print(not_text)
    # print(is_palindrome_str(text))
    # print(is_palindrome_str(not_text))

    functions = [
        (is_palindrome_str, "is_palindrome_str"),
        # (is_palindrome_dict, "is_palindrome_dict"),
        (is_palindrome_str_refactored, "is_palindrome_str_refactored"),
        (quick_check, "quick_check"),
    ]
    functions = functions[::-1]
    duration_list = {}

    # palindromes...
    for func, name in functions:

        # assert func(text) == True

        durations = timeit.repeat(lambda: func(text), repeat=repeat, number=n)
        duration_list[name] = durations
        print(
            "{func:<20}: "
            "min: {min:0.3f}s, mean: {mean:0.3f}s, max: {max:0.3f}s".format(
                func=name,
                min=min(durations),
                mean=np.mean(durations),
                max=max(durations),
            )
        )

    # non palindromes ...
    for func, name in functions:

        assert func(not_text) == False

        durations = timeit.repeat(lambda: func(not_text), repeat=repeat, number=n)
        duration_list[name] = durations
        print(
            "{func:<20}: "
            "min: {min:0.3f}s, mean: {mean:0.3f}s, max: {max:0.3f}s".format(
                func=name,
                min=min(durations),
                mean=np.mean(durations),
                max=max(durations),
            )
        )
        create_boxplot(duration_list)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
""" basefile.py """
# copyright (c) 2019 Michael Treanor
# https://www.github.com/skeptycal
# https://www.twitter.com/skeptycal

# 'Standard Library'
import os
import shutil

from dataclasses import dataclass, field
from pathlib import Path
from pathlib.sys import stat_result

# 'package imports'
from autosys.exceptions.exceptions import BaseFileError


@dataclass
class BaseFile:
    file_name: str
    _path: Path = field(init=False)

    def __post_init__(self):
        try:
            self._path = Path(self.file_name).resolve()
        except:
            raise BaseFileError(
                f"Unable to initialize file '{self.file_name}' ...")

    @property
    def basename(self) -> (str):
        """ Return only the basename of the file. """
        try:
            return self.file_name.split(".")[0].split("/")[-1]
        except:
            raise BaseFileError(f"Unable to return basename of file.")

    @property
    def extension(self) -> (str):
        """ Return only the extension of the file. """
        try:
            return self.file_name.split(".")[-1]
        except:
            raise BaseFileError(f"Unable to return extension of file.")

    @property
    def path(self) -> (Path):
        if not self._path:
            self._path = Path(self.file_name).resolve()
        return self._path

    @property
    def as_posix(self) -> (str):
        return self._path.as_posix()

    @property
    def stat(self) -> (stat_result):
        """ Return os.stat type information.

            e.g.
            os.stat_result(
                st_mode     =33188,
                st_ino      =60176618,
                st_dev      =16777224,
                st_nlink    =1,
                st_uid      =501,
                st_gid      =20,
                st_size     =71,
                st_atime    =1591045258,
                st_mtime    =1591044549,
                st_ctime    =1591044549
                )
            """
        return self.path.stat()


class TextFile(BaseFile):
    def get_line_count(self, filename) -> (int):
        with open(filename) as f:
            return sum(True for line in f)


if False:
    _test_file = "/Users/skeptycal/Documents/coding/python/autosys/autosys/tests/test_basefile.txt"
    test_file = BaseFile(_test_file)
    print(test_file.stat)
    print(os.stat(test_file.path))
    print(test_file.stat == os.stat(test_file.path))
